###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This file contains a class to compute the crystalfield
   coefficients from convoluting the charge density with the potential
   which produces the crystalfield. This is both compatible with
   the Yttrium-Analogue approximation and self-consitent calculation of
   the potential

"""
#TODO: Replace double underscore methods for reading
#TODO: Refactor plotting methods to use general plot_methods

from __future__ import annotations

import h5py
import os
import csv
import logging
import tabulate
from collections import defaultdict
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from pathlib import Path
from scipy.interpolate import interp1d
from scipy.special import sph_harm  #pylint: disable=no-name-in-module
from masci_tools.util.constants import HTR_TO_KELVIN
from masci_tools.io.common_functions import skipHeader
from masci_tools.util.typing import FileLike

from masci_tools.vis.plot_methods import plot_params as mpl_plotter
from masci_tools.vis.parameters import ensure_plotter_consistency, NestedPlotParameters

from typing import NamedTuple, Any
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  #type: ignore[misc]
import warnings

logger = logging.getLogger(__name__)


#This namedtuple is used as the return value for the crystal field calculation to have easy access
#to all the necessary information
class CFCoefficient(NamedTuple):
    """
    Namedtuple representing an individual crystal field coefficient
    """
    l: int
    m: int
    spin_up: float | complex
    spin_down: float | complex
    unit: str
    convention: str


class CFCalculation:
    r"""Class for calculating Crystal Field coefficients using the procedure
    described in C.E. Patrick, J.B. Staunton: J. Phys.: Condens. Matter 31, 305901 (2019)

    Using the formula:
        .. math::
            B_{lm} = \sqrt{\frac{2l+1}{4\pi}} \int^{R_{MT}}\! dr r^2 V_{lm}(r)n_{4f}(r)

        The read in quantities are interpolated from logarithmic meshes to equidistant meshes

        The function constructs an equidistant mesh between 0 and the muffin tin radius
        defined in `self.reference_radius` and with `self.radial_points` points

    :param radial_points: int, number of radial points in the interpolated mesh
    :param reference_radius: stror float; Either 'pot' or 'cdn' or explicit number. Defines which muffin-tin radius
                                is used for the equidistant mesh.
                                IMPORTANT! If txt files are used the muffin-tin radius has to be provided explicitly
    :param coefficient_cutoff: float Defines minimum value that cf coefficients have to have to be considered non-zero

    """

    __version__ = '0.2.0'

    #prefactor for converting Blm to Alm<r^l>
    _alphalm = {
        (0, 0): 1.0,
        (2, 0): 0.5,
        (2, 2): np.sqrt(6.0) / 2.0,
        (4, 0): 1.0 / 8.0,
        (4, 2): np.sqrt(10.0) / 4.0,
        (4, 3): -np.sqrt(35.0) / 2.0,
        (4, 4): np.sqrt(70.0) / 8.0,
        (6, 0): 1.0 / 16.0,
        (6, 2): np.sqrt(105.0) / 16.0,
        (6, 3): -np.sqrt(105.0) / 8.0,
        (6, 4): 3.0 * np.sqrt(14.0) / 16.0,
        (6, 6): np.sqrt(231.0) / 16.0,
    }

    def __init__(self,
                 *,
                 radial_points: int = 4000,
                 reference_radius: float | Literal['pot', 'cdn'] = 'pot',
                 coefficient_cutoff: float | None = 1e-3,
                 **kwargs: Any) -> None:

        self.vlm = {}
        self.cdn = {}

        self.density_normalization: float | None = None
        self.phi: float | None = None
        self.theta: float | None = None

        self.interpolated: bool = False
        self.int = {}
        self.bravaisMat = {}
        self.atom_type: int | None = None

        self.radial_points = radial_points
        self.reference_radius = reference_radius
        self.coefficient_cutoff: float | None = coefficient_cutoff
        self.pot_cutoff = None
        if 'pot_cutoff' in kwargs:
            warnings.warn('The argument pot_cutoff is deprecated. Use cf_cutoff instead', DeprecationWarning)
            self.pot_cutoff = kwargs['pot_cutoff']
        self.only_m0 = False
        if 'only_m0' in kwargs:
            warnings.warn('The argument only_m0 is deprecated.', DeprecationWarning)
            self.only_m0 = kwargs['only_m0']
        if 'quiet' in kwargs:
            warnings.warn('The argument quiet is deprecated.', DeprecationWarning)

    @property
    def denNorm(self):
        """Returns the density normalization

        DEPRECATED: Use density_normalization instead
        """
        return self.density_normalization

    def stevens_prefactor(self, l: int, m: int) -> float:
        """Gives the lm dependent prefactor for conversion between
        Blm and Alm coefficients

        Args:
            :param l: int; orbital quantum number
            :param m: int; magnetic quantum number

        :returns: float prefactor for conversion to Steven's Coefficients
        """
        return self._alphalm.get((l, abs(m)), 0.0)

    def readPot(self, *args, **kwargs):
        """DEPRECATED: Use read_pot"""
        warnings.warn('readPot is deprecated. Use read_potential instead', DeprecationWarning)
        if 'atomType' in kwargs:
            warnings.warn('The argument atomType is deprecated. Use atom_type instead', DeprecationWarning)
            kwargs['atom_type'] = kwargs.pop('atomType')
        if 'complexData' in kwargs:
            warnings.warn('The argument complexData is deprecated. Use complex_data instead', DeprecationWarning)
            kwargs['complex_data'] = kwargs.pop('complexData')
        if 'lm' in kwargs:
            warnings.warn('The argument lm is deprecated. Use lm_indices instead', DeprecationWarning)
            kwargs['lm_indices'] = kwargs.pop('lm')

        self.read_potential(*args, **kwargs)

    def read_potential(self,
                       *files: FileLike | h5py.File,
                       lm_indices: list[tuple[int, int]] | None = None,
                       atom_type: int | None = None,
                       header: int = 0,
                       complex_data: bool = True) -> None:
        """Reads in the potentials for the CF coefficient calculation
        If hdf files are given also the muffin tin radius is read in

        :param args: Expects string filenames for the potentials to read in
                     The function expects either HDF files or txt files with the
                     format (rmesh,vlmup,vlmdn)
        :param lm_indices: list of tuples, Defines the l and m indices for the given txt files
        :param atom_type: int, Defines the atomType to read in (only for HDF files)
        :param header: int, Define how many lines to skip in the beginning of txt file
        :param complex_data: bool, Define if the data in the text file is complex

        Raises:
            ValueError: lm indices list length has to match number of files read in

        """

        if lm_indices is None:
            lm_indices = []

        #Reads in the filenames given in args as potentials
        for index, file in enumerate(files):
            if isinstance(file, (str, Path)):
                _, extension = os.path.splitext(file)
                if extension == '.hdf':
                    with h5py.File(file, 'r') as hdffile:
                        self.__readpotHDF(hdffile, atom_type=atom_type)
                else:
                    if index >= len(lm_indices):
                        raise ValueError('Not enough lm indices for the given files')
                    self.__readpottxt(file, lm_indices[index], header=header, complexData=complex_data)
            else:
                self.__readpotHDF(file, atom_type)

    @classmethod
    def from_arrays(cls, charge_density: np.ndarray, potentials: dict[tuple[int, int], np.ndarray],
                    radial_mesh: dict[str, np.ndarray], **kwargs: Any) -> CFCalculation:
        """
        Create a CFCalculation instance from arrays

        .. warning::
            Using this classmethod will not check that the charge density and potential have the same bravais
            matrix

        :param charge_density: Array for the normed charge density
        :param potentials: Data for the potentials (dict mapping (l,m) tuples to the corresponding data)
        :param radial_mesh: dict for the data for the used radial meshs (keys 'cdn' and 'pot')

        Other Kwargs are passed on to the constructor
        """

        cfcalc = cls(**kwargs)

        cfcalc.cdn['data'] = charge_density
        cfcalc.cdn['rmesh'] = radial_mesh['cdn']
        cfcalc.cdn['RMT'] = np.max(radial_mesh['cdn'])

        cfcalc.vlm = potentials
        cfcalc.vlm['rmesh'] = radial_mesh['pot']
        cfcalc.vlm['RMT'] = np.max(radial_mesh['pot'])

        return cfcalc

    def readCDN(self, *args, **kwargs):
        """DEPRECATED: Use read_charge_density instead"""
        warnings.warn('readCDN is deprecated. Use read_charge_density instead', DeprecationWarning)
        if 'atomType' in kwargs:
            warnings.warn('The argument atomType is deprecated. Use atom_type instead', DeprecationWarning)
            kwargs['atom_type'] = kwargs.pop('atomType')
        self.read_charge_density(*args, **kwargs)

    def read_charge_density(self, file: FileLike | h5py.File, atom_type: int | None = None, header: int = 0) -> None:
        """Reads in the normed charge density for the CF coefficient calculation
        If hdf files are given also the muffin tin radius is read in

        Parameters:
            :param file: Expects string filename for the charge density to read in
                         The function expects either HDF files or txt files with the
                         format (rmesh,cdn).
                         The charge density should be given as r^2n(r) and normed to 1
        kwargs:
            :param atom_type: int, Defines the atom_type to read in (only for HDF files)
            :param header: int, Define how many lines to skip in the beginning of txt file

        """
        if isinstance(file, (str, Path)):
            _, extension = os.path.splitext(file)
            if extension == '.hdf':
                with h5py.File(file, 'r') as hdffile:
                    self.__readcdnHDF(hdffile, atom_type=atom_type)
            else:
                self.__readcdntxt(file, header=header)
        else:
            self.__readcdnHDF(file, atom_type)

    def __readpotHDF(self, hdffile, atom_type=None):
        """Read in the potential from a HDF file

        """

        info = hdffile.get('general')
        numPOT = info.attrs['numPOT'][0]
        if 'bravaisMatrix' in info:
            self.bravaisMat['pot'] = np.array(info.get('bravaisMatrix'))

        if numPOT == 0:
            raise ValueError(f'No potentials found in {hdffile}')

        potential_groups = {key for key in hdffile if 'pot-' in key}

        if len(potential_groups) != 1 and atom_type is None:
            raise ValueError('Multiple possibilities for calculated potentials. '
                             f'Select the desired atomType: {potential_groups}')

        if atom_type is not None:
            pot_group = f'pot-{atom_type}'
        else:
            pot_group = potential_groups.pop()
            _, _, atom_type = pot_group.partition('-')

        if self.atom_type is not None and self.atom_type != atom_type:
            logger.warning('atom_type %s for potential does not correspond to previously read in data (%s)', atom_type,
                           self.atom_type)

        if self.atom_type is None:
            self.atom_type = atom_type

        if pot_group in hdffile:
            _pot = hdffile.get(pot_group)

            self.vlm['RMT'] = _pot.attrs['RMT'][0]
            for key in _pot.keys():

                if key == 'rmesh':
                    _rmesh = _pot.get(key)
                    self.vlm['rmesh'] = np.array(_rmesh)
                else:
                    _vlm = _pot.get(key)
                    l = _vlm.attrs['l'][0]
                    m = _vlm.attrs['m'][0]

                    _data = _vlm.get('vlm')
                    _data = np.array(_data[:, :, 0] + 1j * _data[:, :, 1])
                    if self.pot_cutoff is None or abs(_data).max() >= self.pot_cutoff:
                        self.vlm[(l, m)] = _data

        else:
            raise ValueError(f'No potential for atomType {atom_type} found in {hdffile}')

        logger.info('read_potential (HDF): Generated the following information: %s', self.vlm.keys())

    def __readpottxt(self, file, index, header=0, complexData=True):
        """Read in the potential for the (l,m) tuple 'index' from a txt file
        The muffin-tin radius is inferred from the biggest argument in the rmesh
        """

        with open(file, newline='', encoding='utf-8') as csvfile:
            reader = csv.reader(csvfile, delimiter=' ', quotechar='|')

            if index not in self.vlm:
                self.vlm[index] = []
                self.vlm[index].append([])
                self.vlm[index].append([])
            else:
                raise KeyError(f'Multiple definitions for potential {index}')

            self.vlm['rmesh'] = []

            for line in skipHeader(reader, header):
                while '' in line:
                    line.remove('')
                self.vlm['rmesh'].append(float(line[0]))
                if complexData:
                    self.vlm[index][0].append(float(line[1]) + 1j * float(line[2]))
                    self.vlm[index][1].append(float(line[3]) + 1j * float(line[4]))
                else:
                    self.vlm[index][0].append(float(line[1]))
                    self.vlm[index][1].append(float(line[2]))

            self.vlm[index] = np.array(self.vlm[index])
            self.vlm['rmesh'] = np.array(self.vlm['rmesh'])
            self.vlm['RMT'] = max(self.vlm['rmesh'])

    def __readcdnHDF(self, hdffile, atom_type=None):
        """Read in the charge density from a HDF file

        """

        info = hdffile.get('general')
        numCDN = info.attrs['numCDN'][0]
        if 'bravaisMatrix' in info:
            self.bravaisMat['cdn'] = np.array(info.get('bravaisMatrix'))

        if numCDN == 0:
            raise ValueError(f'No charge densities found in {hdffile}')

        cdn_groups = {key for key in hdffile if 'cdn-' in key}

        if len(cdn_groups) != 1 and atom_type is None:
            raise ValueError('Multiple possibilities for calculated charge densities. '
                             f'Select the desired atom_type: {cdn_groups}')

        if atom_type is not None:
            cdn_group = f'cdn-{atom_type}'
        else:
            cdn_group = cdn_groups.pop()
            _, _, atom_type = cdn_group.partition('-')

        if self.atom_type is not None and self.atom_type != atom_type:
            logger.warning('atom_type %s for density does not correspond to previously read in data (%s)', atom_type,
                           self.atom_type)

        if self.atom_type is None:
            self.atom_type = atom_type

        if cdn_group in hdffile:
            _cdn = hdffile.get(cdn_group)
            self.cdn['RMT'] = _cdn.attrs['RMT'][0]
            _rmesh = _cdn.get('rmesh')
            self.cdn['rmesh'] = np.array(_rmesh)
            _data = _cdn.get('cdn')
            self.cdn['data'] = np.array(_data)

        else:
            raise ValueError(f'No charge density for atom_type {atom_type} found in {hdffile}')

        logger.info('read_charge_density (HDF): Generated the following information: %s', self.cdn.keys())

    def __readcdntxt(self, file, header=0):
        """Read in the charge density from a txt file
        The muffin-tin radius is inferred from the biggest argument in the rmesh
        """

        with open(file, newline='', encoding='utf-8') as csvfile:
            reader = csv.reader(csvfile, delimiter=' ', quotechar='|')

            if 'data' not in self.cdn:
                self.cdn['data'] = []
            else:
                raise KeyError('Charge density already read in')

            self.cdn['rmesh'] = []

            for line in skipHeader(reader, header):
                while '' in line:
                    line.remove('')

                self.cdn['rmesh'].append(float(line[0]))
                self.cdn['data'].append(float(line[1]))

            self.cdn['data'] = np.array(self.cdn['data'])
            self.cdn['rmesh'] = np.array(self.cdn['rmesh'])
            self.cdn['RMT'] = max(self.cdn['rmesh'])

    def validate(self) -> None:
        """Validate that the object can be used to execute the calculation
        Checks that the given bravais matrices are equal if given
        """

        if 'data' not in self.cdn or 'rmesh' not in self.cdn:
            raise ValueError('Charge density input incomplete')

        if 'rmesh' not in self.vlm or len(self.vlm.items()) <= 2:
            raise ValueError('Potential input incomplete')

        if 'cdn' in self.bravaisMat and 'pot' in self.bravaisMat:
            diffBravais = self.bravaisMat['cdn'] - self.bravaisMat['pot']
            max_diff = np.max(np.abs(diffBravais))
            if max_diff > 1e-8:
                logger.warning('Differing definitions of potentials and charge density bravais matrix. Max. Diff: %f',
                               max_diff)

    def interpolate(self) -> None:
        """Interpolate all quantities to a common equidistant radial mesh

        """

        if self.reference_radius == 'pot':
            refRMT = self.vlm['RMT']
        elif self.reference_radius == 'cdn':
            refRMT = self.cdn['RMT']
        else:
            refRMT = self.reference_radius

        self.int = defaultdict(dict)

        self.int['rmesh'] = np.arange(0.0, refRMT, refRMT / self.radial_points)
        self.int['cdn'] = interp1d(self.cdn['rmesh'], self.cdn['data'], fill_value='extrapolate')

        for key, value in self.vlm.items():
            if key not in ('RMT', 'rmesh'):
                self.int[key]['spin-up'] = interp1d(self.vlm['rmesh'], value[0, :], fill_value='extrapolate')
                if value.shape[0] == 2:
                    self.int[key]['spin-down'] = interp1d(self.vlm['rmesh'], value[1, :], fill_value='extrapolate')
        self.interpolated = True

    def get_coefficients(self,
                         convention: Literal['Stevens', 'Wybourne'] = 'Stevens',
                         table_fmt: str | None = None) -> list[CFCoefficient]:
        """Performs the integration to obtain the crystal field coefficients
        If the data was not already interpolated, the interpolation will
        be performed beforehand

        Parameters:
            :param convention: str of the convention to use (Stevens or Wybourne)
            :param table_fmt: str if not None a table with the given format will be printed using tabulate

        :returns: list of CFCoefficient objects (namedtuple), with all the necessary information
        """

        if convention not in ('Stevens', 'Wybourne'):
            raise ValueError(f'Unknown Crystal field convention: {convention}')

        self.validate()

        if 'cdn' in self.bravaisMat and 'pot' in self.bravaisMat:
            a_vector = self.bravaisMat['pot'][0, :]
            c_vector = self.bravaisMat['pot'][2, :]

            self.theta = np.arccos(c_vector[2] / (np.linalg.norm(c_vector)))
            self.phi = np.arccos(a_vector[0] / (np.linalg.norm(a_vector)))

            logger.info(r'Angle between lattice vector c and z-axis: %f $\pi$', self.theta / np.pi)
            logger.info(r'Angle between lattice vector a and x-axis:  %f $\pi$', self.phi / np.pi)

        if not self.interpolated:
            self.interpolate()

        self.density_normalization = np.trapz(self.int['cdn'](self.int['rmesh']), self.int['rmesh'])
        logger.info('Density normalization = %f', self.density_normalization)

        results = []
        for lmkey, vlm in [(key, val) for key, val in self.int.items() if isinstance(key, tuple)]:
            l, m = lmkey
            if not self.only_m0 or m == 0:
                integral = {}
                for key, pot in vlm.items():
                    integral[key] = np.trapz(
                        pot(self.int['rmesh']) * self.int['cdn'](self.int['rmesh']), self.int['rmesh'])
                    integral[key] *= np.sqrt((2.0 * l + 1.0) / (4.0 * np.pi)) * HTR_TO_KELVIN

                if 'spin-down' not in integral:
                    integral['spin-down'] = integral['spin-up']

                if convention == 'Stevens':
                    integral = {key: val.real * self.stevens_prefactor(l, m) for key, val in integral.items()}

                if self.coefficient_cutoff is not None:
                    if all(np.abs(value) < self.coefficient_cutoff for value in integral.values()):
                        logger.info('Dismissing coefficient for (%i,%i): %s', l, m, integral)
                        continue

                results.append(
                    CFCoefficient(l=l,
                                  m=m,
                                  spin_up=integral['spin-up'],
                                  spin_down=integral['spin-down'],
                                  unit='K',
                                  convention=convention))

        results.sort(key=lambda item: (item.l, abs(item.m)))

        if table_fmt is not None:

            print(f'\nThe following results were obtained with the {convention} convention:')

            results_table = [
                [val for key, val in entry._asdict().items() if key not in ('unit', 'convention')] for entry in results
            ]

            headers = ('l', 'm', 'Spin-Up [K]', 'Spin-Down [K]')
            if table_fmt in ('latex', 'latex_raw'):
                table_fmt = 'latex_raw'
                if convention == 'Stevens':
                    headers = ('l', 'm', r'$A^{\uparrow}_{lm}\langle r^l \rangle$ [K]',
                               r'$A^{\downarrow}_{lm}\langle r^l \rangle$ [K]')
                else:
                    headers = ('l', 'm', r'$B^{\uparrow}_{lm}$ [K]', r'$B^{\downarrow}_{lm}$ [K]')

            print(tabulate.tabulate(results_table, tablefmt=table_fmt, headers=headers))

        return results

    def performIntegration(self, convert=True):
        """DEPRECATED: Use get_coefficients instead

        Performs the integration to obtain the crystal field coefficients
        If the data was not already interpolated, the interpolation will
        be performed beforehand

        Parameters:
            :param convert: bool, converts to Steven's coefficients (if True)

        :returns: list of CFCoefficient objects (namedtuple), with all the necessary information

        """
        warnings.warn('performIntegration is deprecated. Use get_coefficients instead', DeprecationWarning)
        return self.get_coefficients(convention='Stevens' if convert else 'Wybourne')

    @property
    def nonzero_coefficients_lm(self) -> list[tuple[int, int]]:
        """
        Return the lm indices of coefficients that are bigger than the set cutoff
        """
        if self.coefficient_cutoff is None:
            warnings.warn('Coefficient cutoff not set. All coefficients are returned', UserWarning)
        return [(coeff.l, coeff.m) for coeff in self.get_coefficients()]

    @property
    def spin_polarized(self) -> bool:
        """
        Return whether the potentials have a spin-down component
        """
        return any(entry.shape[0] == 2 for key, entry in self.vlm.items() if key not in ('RMT', 'rmesh'))

    def get_potentials(self,
                       spin: Literal['up', 'down'],
                       interpolated: bool = True,
                       only_nonzero: bool = True,
                       complex_data: bool = True) -> tuple[np.ndarray, dict[tuple[int, int], np.ndarray]]:
        """
        Return the potentials and the corresponding radial mesh

        :param spin: which spin to return
        :param interpolated: bool, if True the interpolated mesh and potentials are returned
        :param only_nonzero: bool, if True only the potentials corresponding to a non-zero coefficient are returned
        :param complex_data: bool, if False only the real part of the potentials is returned
        """
        if only_nonzero:
            indices = self.nonzero_coefficients_lm
        else:
            indices = [key for key in self.vlm.keys() if key not in ('RMT', 'rmesh')]

        if interpolated:
            rmesh = self.int['rmesh']
            potentials = {key: self.int[key][f'spin-{spin}'](rmesh) for key in indices}
        else:
            rmesh = self.vlm['rmesh']
            spin_index = 0 if spin == 'up' else 1
            potentials = {key: self.vlm[key][spin_index] for key in indices}

        if not complex_data:
            potentials = {key: pot.real for key, pot in potentials.items()}

        return rmesh, potentials

    def get_charge_density(self, interpolated=True) -> tuple[np.ndarray, np.ndarray]:
        """
        Return the charge density and the corresponding radial mesh

        :param spin: which spin to return
        :param interpolated: bool, if True the interpolated mesh and density are returned
        """

        if interpolated:
            rmesh = self.int['rmesh']
            density = self.int['cdn'](rmesh)
        else:
            rmesh = self.vlm['rmesh']
            density = self.cdn['data']

        return rmesh, density


@ensure_plotter_consistency(mpl_plotter)
def plot_crystal_field_calculation(cfcalc,
                                   *,
                                   saveas='crystal_field_calc',
                                   potential_title='Potential',
                                   density_title='Density',
                                   xlabel='$R$ (Bohr)',
                                   potential_ylabel='$Vpot$ (Hartree)',
                                   density_ylabel='Density',
                                   density_kwargs=None,
                                   potential=True,
                                   density=True,
                                   axis_potential=None,
                                   axis_density=None,
                                   **kwargs):
    """
    Plot the given potentials and charge densities used in the given
    :py:class:`CFCalculation`

    :param cfcalc: CFcalculation containing the data to plot
    :param saveas: str, Define the filename to save the figure
    :param potential_title: Title for the potential subplot
    :param density_title: Title for the charge density subplot
    :param xlabel: label for the x axis of both subplots
    :param potential_ylabel: label for the y axis of the potential subplot
    :param density_ylabel: label for the y axis f the charge density subplot
    :param density_kwargs: dict with keyword argument passed to the plotting of the
                           density
    :param figure: Use this preexisting matplotlib figure to produce the plots

    All other Kwargs are passed on to the :py:func:`~masci_tools.vis.plot_methods.multiple_scatterplots()`
    calls for plotting the potentials
    """
    from masci_tools.vis.plot_methods import multiple_scatterplots

    if density_kwargs is None:
        density_kwargs = {}

    if 'filename' in kwargs:
        warnings.warn('filename is deprecated. Use saveas instead', DeprecationWarning)
        saveas = kwargs.pop('filename')

    if 'pot_ylabel' in kwargs:
        warnings.warn('pot_ylabel is deprecated. Use potential_ylabel instead', DeprecationWarning)
        potential_ylabel = kwargs.pop('pot_ylabel')

    if 'cdn_ylabel' in kwargs:
        warnings.warn('cdn_ylabel is deprecated. Use density_ylabel instead', DeprecationWarning)
        density_ylabel = kwargs.pop('cdn_ylabel')

    if 'pot_title' in kwargs:
        warnings.warn('pot_title is deprecated. Use potential_title instead', DeprecationWarning)
        potential_title = kwargs.pop('pot_title')

    if 'cdn_title' in kwargs:
        warnings.warn('cdn_title is deprecated. Use density_title instead', DeprecationWarning)
        density_title = kwargs.pop('cdn_title')

    if 'save' in kwargs:
        warnings.warn('save is deprecated. Use save_plots instead', DeprecationWarning)
        kwargs['save_plots'] = kwargs.pop('save')

    rmesh, potentials = cfcalc.get_potentials('up', complex_data=False)

    indices, ydata = zip(*sorted(potentials.items(), key=lambda x: x[0]))
    labels = [rf'$V_{{{l}{m}}}$' for l, m in indices]

    figure = None
    if potential and density and axis_potential is None and axis_density is None:
        figure = plt.figure()
        gs = figure.add_gridspec(1, 2)
        axis_potential = figure.add_subplot(gs[0])
        axis_density = figure.add_subplot(gs[1])

    mpl_plotter.single_plot = False
    mpl_plotter.num_plots = len(indices)

    mpl_plotter.set_defaults(
        default_type='function',
        plot_label=labels,
        legend=True,
        legend_options={
            'loc': 'upper left',
            'ncol': 1,
            'borderaxespad': 0.0,
            'linewidth': None,
        },
        marker=None,
        color_cycle=['black', 'red', 'blue', 'orange', 'green', 'purple'],
    )

    save_keys = {'show', 'save_plots', 'save_format', 'save_options'}
    save_options = {key: val for key, val in kwargs.items() if key in save_keys}
    kwargs = {key: val for key, val in kwargs.items() if key not in save_keys}

    mpl_plotter.set_parameters(**save_options)

    if potential:
        with NestedPlotParameters(mpl_plotter):
            axis_potential = multiple_scatterplots(rmesh,
                                                   ydata,
                                                   xlabel=xlabel,
                                                   ylabel=potential_ylabel,
                                                   title=potential_title,
                                                   show=False,
                                                   save_plots=False,
                                                   axis=axis_potential,
                                                   **kwargs)

        mpl_plotter.set_defaults(default_type='function', plot_label=None, linestyle='--', legend=False)

        if cfcalc.spin_polarized:
            rmesh, potentials = cfcalc.get_potentials('down', complex_data=False)

            _, ydata = zip(*sorted(potentials.items(), key=lambda x: x[0]))
            with NestedPlotParameters(mpl_plotter):
                axis_potential = multiple_scatterplots(rmesh,
                                                       ydata,
                                                       xlabel=xlabel,
                                                       ylabel=potential_ylabel,
                                                       title=potential_title,
                                                       show=False,
                                                       save_plots=False,
                                                       axis=axis_potential,
                                                       **kwargs)

    rmesh_int, density_int = cfcalc.get_charge_density()
    rmesh, charge_density = cfcalc.get_charge_density(interpolated=False)

    xdata = [rmesh, rmesh_int]
    ydata = [charge_density, density_int]

    if density:
        mpl_plotter.single_plot = False
        mpl_plotter.num_plots = 2

        mpl_plotter.set_defaults(default_type='function',
                                 plot_label=[r'$n(r)$', None],
                                 linestyle=['-', '--'],
                                 color=['black', 'blue'],
                                 legend=True)

        with NestedPlotParameters(mpl_plotter):
            axis_density = multiple_scatterplots(xdata,
                                                 ydata,
                                                 xlabel=xlabel,
                                                 ylabel=density_ylabel,
                                                 title=density_title,
                                                 show=False,
                                                 save_plots=False,
                                                 axis=axis_density,
                                                 **density_kwargs)

    if figure is not None:
        figure.set_size_inches(14.0, 10.0)
        figure.subplots_adjust(left=0.10, bottom=0.2, right=0.90, wspace=0.4, hspace=0.4)

    mpl_plotter.save_plot(saveas)

    return [axis_potential, axis_density]


@ensure_plotter_consistency(mpl_plotter)
def plot_crystal_field_potential(cfcoeffs,
                                 *,
                                 saveas='crystal_field_potential_areaplot',
                                 spin='avg',
                                 phi=0.0,
                                 figure=None,
                                 **kwargs):
    """
    Plots the angular dependence of the calculated CF potential as well
    as a plane defined by phi.

    :param cfcoeffs: list of CFCoefficients to construct the potential
    :param saveas: str, defines the filename to save the figure
    :param spin: str; Either 'up', 'dn' or 'avg'. Which spin direction to plot
                    ('avg'-> ('up'+'dn')/2.0)
    :param phi: float, defines the phi angle of the plane

    :raises AssertionError: When coefficients are provided as wrong types or in the wrong convention
    :raises ValueError: When coefficients are provided in the wrong convention

    """
    from masci_tools.vis.plot_methods import colormesh_plot

    assert all(isinstance(coeff, CFCoefficient) for coeff in cfcoeffs), \
           'Only provide a list of CFCoefficients to plot_crystal_field_potential'

    if 'filename' in kwargs:
        warnings.warn('filename is deprecated. Use saveas instead', DeprecationWarning)
        saveas = kwargs.pop('filename')

    if 'save' in kwargs:
        warnings.warn('save is deprecated. Use save_plots instead', DeprecationWarning)
        kwargs['save_plots'] = kwargs.pop('save')

    if spin not in ('up', 'down', 'avg'):
        raise ValueError(f"Invalid Argument for spin: '{spin}'")

    if not all(coeff.convention == 'Wybourne' for coeff in cfcoeffs):
        raise ValueError('Wrong convention for plotting in spherical harmonics basis')

    #generate the theta/phi meshgrid
    theta_grid = np.linspace(0, np.pi, 181)
    phi_grid = np.linspace(0.0, 2.0 * np.pi, 361)
    xv, yv = np.meshgrid(phi_grid, theta_grid)
    cf_grid = np.zeros((181, 361), dtype='complex')

    for coeff in cfcoeffs:
        if spin == 'avg':
            value = 0.5 * (coeff.spin_up + coeff.spin_down)
        elif spin == 'up':
            value = coeff.spin_up
        elif spin == 'down':
            value = coeff.spin_down
        if coeff.m >= 0:
            cf_grid += value * sph_harm(coeff.m, coeff.l, xv, yv)

    #Plot the angular dependence
    maxv = max(cf_grid.real.max(), abs(cf_grid.real.min()))

    if figure is None:
        figure = plt.figure(figsize=(15, 5))

    gs = mpl.gridspec.GridSpec(1, 2, width_ratios=[2.0, 1.5])
    ax = plt.subplot(gs[0])

    mpl_plotter.set_defaults(default_type='function',
                             cmap='coolwarm',
                             limits={'color': (-maxv, maxv)},
                             xticks=[0.0, np.pi / 2.0, np.pi, 3 * np.pi / 2.0, 2 * np.pi],
                             xticklabels=[r'0', r'$\pi/2$', r'$\pi$', r'$3\pi/2$', r'$2\pi$'],
                             yticks=[0.0, np.pi / 4.0, np.pi / 2.0, 3 * np.pi / 4.0, np.pi],
                             yticklabels=[r'$\pi/2$', r'$\pi/4$', r'$0.0$', r'$-\pi/4$', r'$-\pi/2$'])

    outer_keys = {'show', 'save_plots', 'save_format', 'save_options', 'colorbar'}
    outer_options = {key: val for key, val in kwargs.items() if key in outer_keys}
    kwargs = {key: val for key, val in kwargs.items() if key not in outer_keys}

    mpl_plotter.set_parameters(**outer_options)

    #Angular dependence plot

    gs = mpl.gridspec.GridSpec(1, 2, width_ratios=[2.0, 1.5])
    ax = plt.subplot(gs[0])
    with NestedPlotParameters(mpl_plotter):
        ax = colormesh_plot(xv,
                            yv,
                            cf_grid.real,
                            xlabel=r'$\phi$',
                            ylabel=r'$\theta$',
                            title='Angular Dependence',
                            axis=ax,
                            show=False,
                            save_plots=False,
                            colorbar=False,
                            vmin=-maxv,
                            vmax=maxv,
                            **kwargs)

    if np.abs(phi_grid - phi).min() > 1e-5:
        raise ValueError(f'Angle {phi} not found in grid')
    phi_ind = np.abs(phi_grid - phi).argmin()

    theta_grid_pm = list(-1.0 * theta_grid)
    theta_cf = list(cf_grid[:, phi_ind])
    for index, theta in enumerate(theta_grid):
        theta_grid_pm.append(theta)
        theta_cf.append(theta_cf[index])

    theta_grid = np.array(theta_grid)
    theta_cf = np.array(theta_cf)

    #interpolate
    theta_cf = interp1d(theta_grid_pm, theta_cf.real, fill_value='extrapolate')

    nx = 200
    ny = 200
    #Define the cartesian grid between -1 and 1
    x = np.linspace(-1, 1, nx)
    y = np.linspace(-1, 1, ny)
    xv, yv = np.meshgrid(x, y)
    z = theta_cf(np.arctan(xv / yv))

    phi_fract = phi / np.pi

    ax = plt.subplot(gs[1])
    mpl_plotter.set_defaults(default_type='function',
                             cmap='coolwarm',
                             limits={'color': (-maxv, maxv)},
                             xticks=[-1.0, -0.5, 0.0, 0.5, 1.0],
                             yticks=[-1.0, -0.5, 0.0, 0.5, 1.0],
                             colorbar_options={
                                 'shrink': 0.8,
                                 'label': r'$V_{{CF}}$ [K]',
                                 'labelsize': 14
                             })

    with NestedPlotParameters(mpl_plotter):
        ax = colormesh_plot(xv,
                            yv,
                            z,
                            xlabel=r'x [Bohr]',
                            ylabel=r'y [Bohr]',
                            title=rf'Crystal Field potential for $\phi={{{phi_fract:.2f}}}\pi$',
                            axis=ax,
                            show=False,
                            save_plots=False,
                            colorbar=False,
                            vmin=-maxv,
                            vmax=maxv,
                            **kwargs)

    mpl_plotter.show_colorbar(ax)
    figure.set_constrained_layout_pads(w_pad=0., h_pad=0.0, hspace=0., wspace=0.)

    mpl_plotter.save_plot(saveas)

    return figure



"""
Commands for parsing information from KKR/Fleur files
"""
from .root import cli

import click
from masci_tools.cmdline.utils import echo


@cli.group('parse')
def parse():
    """Commands for parsing information from KKR/Fleur files"""


def _load_xml_file(xml_file):
    from masci_tools.io.fleur_xml import load_inpxml, load_outxml
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.io.parsers.fleur_schema import InputSchemaDict
    from lxml import etree

    FALLBACK_VERSION = '0.34'

    try:
        xmltree, schema_dict = load_outxml(xml_file)
    except (ValueError, IndexError):
        try:
            xmltree, schema_dict = load_inpxml(xml_file)
        except ValueError:
            schema_dict = InputSchemaDict.fromVersion(FALLBACK_VERSION)
            parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
            xmltree = etree.parse(xml_file, parser)
            xmltree, _ = clear_xml(xmltree)

    return xmltree, schema_dict


@parse.command('inp-file')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_inp_file(xml_file):
    """
    Parse the Fleur inp.xml into a python dictionary
    """
    from masci_tools.io.parsers.fleur import inpxml_parser

    parser_info = {}
    parser_dict = inpxml_parser(xml_file, parser_info_out=parser_info)

    echo.echo_info('Parser output:')
    echo.echo_dictionary(parser_dict)
    echo.echo_info('Parser warnings/information:')
    echo.echo_dictionary(parser_info)


@parse.command('out-file')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--ignore-validation', is_flag=True)
def parse_out_file(xml_file, ignore_validation):
    """
    Parse the Fleur out.xml into a python dictionary
    """
    from masci_tools.io.parsers.fleur import outxml_parser

    parser_info = {}
    parser_dict = outxml_parser(xml_file, parser_info_out=parser_info, ignore_validation=ignore_validation)

    echo.echo_info('Parser output:')
    echo.echo_dictionary(parser_dict)
    echo.echo_info('Parser warnings/information:')
    echo.echo_dictionary(parser_info)


@parse.command('constants')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_constants(xml_file):
    """
    Parse the mathematical constants used in the given xml-file
    """
    from masci_tools.io.fleur_xml import get_constants

    xmltree, schema_dict = _load_xml_file(xml_file)

    constants = get_constants(xmltree, schema_dict)
    echo.echo_dictionary(constants)


@parse.command('fleur-modes')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_fleur_modes(xml_file):
    """
    Parse the Fleur modes of the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_fleur_modes

    xmltree, schema_dict = _load_xml_file(xml_file)

    fleur_modes = get_fleur_modes(xmltree, schema_dict)
    echo.echo_dictionary(fleur_modes)


@parse.command('structure')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_structure_data(xml_file):
    """
    Parse the structure information in the given Fleur xml file
    """
    from masci_tools.util.xml.xml_getters import get_structure_data

    xmltree, schema_dict = _load_xml_file(xml_file)

    atoms, cell, pbc = get_structure_data(xmltree, schema_dict)

    echo.echo_info('Atoms found:')
    echo.echo_formatted_list(atoms, ['symbol', 'kind', 'position'])
    echo.echo_info('Bravais matrix:')
    echo.echo(str(cell))
    echo.echo_info(f'Periodic boundary conditions {pbc}:')


@parse.command('cell')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_cell_data(xml_file):
    """
    Parse the unit cell definition of the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_cell

    xmltree, schema_dict = _load_xml_file(xml_file)

    cell, pbc = get_cell(xmltree, schema_dict)

    echo.echo_info('Bravais matrix:')
    echo.echo(str(cell))
    echo.echo_info(f'Periodic boundary conditions {pbc}:')


@parse.command('parameters')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_parameter_data(xml_file):
    """
    Parse the calculation parameters of the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_parameter_data

    xmltree, schema_dict = _load_xml_file(xml_file)

    params = get_parameter_data(xmltree, schema_dict)

    echo.echo_info('LAPW parameters:')
    echo.echo_dictionary(params)


@parse.command('nkpts')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_nkpts(xml_file):
    """
    Extract the number of kpoints used in the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_nkpts

    xmltree, schema_dict = _load_xml_file(xml_file)
    nkpts = get_nkpts(xmltree, schema_dict)

    echo.echo_info(f'Number of k-points: {nkpts}')


@parse.command('kpoints')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_kpoints_data(xml_file):
    """
    Parse the used kpoints from the given xml-file
    """
    from masci_tools.util.xml.xml_getters import get_kpoints_data

    xmltree, schema_dict = _load_xml_file(xml_file)
    kpoints, weights, cell, pbc = get_kpoints_data(xmltree, schema_dict, only_used=True)

    echo.echo_info('Bravais matrix:')
    echo.echo(str(cell))
    echo.echo_info(f'Periodic boundary conditions {pbc}:')
    echo.echo_info('Kpoint coordinates (relative) and weights:')
    for kpoint, weight in zip(kpoints, weights):
        echo.echo(f'{kpoint}    w={weight}')


@parse.command('relaxation')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_relaxation_data(xml_file):
    """
    Parse the relaxation information for the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_relaxation_information

    xmltree, schema_dict = _load_xml_file(xml_file)
    relax_info = get_relaxation_information(xmltree, schema_dict)

    echo.echo_info('Relaxation Information:')
    echo.echo_dictionary(relax_info)


@parse.command('symmetry')
@click.argument('xml-file', type=click.Path(exists=True))
def parse_symmetry_information(xml_file):
    """
    Parse the symmetry information for the given xml file
    """
    from masci_tools.util.xml.xml_getters import get_symmetry_information

    xmltree, schema_dict = _load_xml_file(xml_file)
    rotations, shifts = get_symmetry_information(xmltree, schema_dict)

    echo.echo_info('Symmetry information:')
    for rotation, shift in zip(rotations, shifts):
        echo.echo('Rotation:')
        echo.echo(str(rotation))
        echo.echo('Translation:')
        echo.echo(str(shift))
        echo.echo('')


@parse.command('attrib')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
@click.option('--tag', '-t', type=str)
def parse_attrib(xml_file, name, contains, not_contains, tag):
    """
    Parse the specified attribute from the given xml file
    """
    from masci_tools.util.schema_dict_util import evaluate_attribute

    xmltree, schema_dict = _load_xml_file(xml_file)
    attribv = evaluate_attribute(xmltree, schema_dict, name, contains=contains, not_contains=not_contains, tag_name=tag)

    echo.echo(f'Value for attribute {name}: {attribv}')


@parse.command('text')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
def parse_text(xml_file, name, contains, not_contains):
    """
    Parse the text of the specified tag from the given xml file
    """
    from masci_tools.util.schema_dict_util import evaluate_text

    xmltree, schema_dict = _load_xml_file(xml_file)
    textv = evaluate_text(xmltree, schema_dict, name, contains=contains, not_contains=not_contains)

    echo.echo(f'Text for tag {name}: {textv}')


@parse.command('all-attribs')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
@click.option('--subtags', is_flag=True)
@click.option('--text', is_flag=True)
def parse_all_attribs(xml_file, name, contains, not_contains, subtags, text):
    """
    Parse all attributes of the specified tag from the given xml file
    """
    from masci_tools.util.schema_dict_util import evaluate_tag

    xmltree, schema_dict = _load_xml_file(xml_file)
    res = evaluate_tag(xmltree,
                       schema_dict,
                       name,
                       contains=contains,
                       not_contains=not_contains,
                       subtags=subtags,
                       text=text)

    echo.echo(f'Tag {name}:')
    echo.echo_dictionary(res)


@parse.command('parent-attribs')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
def parse_parent_attribs(xml_file, name, contains, not_contains):
    """
    Parse all attributes of the parent of the specified tag from the given xml file
    """
    from masci_tools.util.schema_dict_util import evaluate_parent_tag

    xmltree, schema_dict = _load_xml_file(xml_file)
    res = evaluate_parent_tag(xmltree, schema_dict, name, contains=contains, not_contains=not_contains)

    echo.echo(f'Tag {name}:')
    echo.echo_dictionary(res)


@parse.command('tag-exists')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
def parse_tag_exists(xml_file, name, contains, not_contains):
    """
    Return whether the specified tag exists in the given xml file
    """
    from masci_tools.util.schema_dict_util import tag_exists

    xmltree, schema_dict = _load_xml_file(xml_file)
    res = tag_exists(xmltree, schema_dict, name, contains=contains, not_contains=not_contains)

    echo.echo(f"Tag {name}: {'exists' if res else 'does not exist'}")


@parse.command('number-nodes')
@click.argument('xml-file', type=click.Path(exists=True))
@click.option('--name', '-n', type=str)
@click.option('--contains', '-c', type=str, multiple=True)
@click.option('--not-contains', '-nc', type=str, multiple=True)
def parse_number_nodes(xml_file, name, contains, not_contains):
    """
    Return how often the specified tag occurs in the given xml file
    """
    from masci_tools.util.schema_dict_util import get_number_of_nodes

    xmltree, schema_dict = _load_xml_file(xml_file)
    res = get_number_of_nodes(xmltree, schema_dict, name, contains=contains, not_contains=not_contains)

    echo.echo(f'Tag {name}: {res} times')



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Load all fleur schema related functions
"""
from .schema_dict import (InputSchemaDict, OutputSchemaDict, SchemaDict, schema_dict_version_dispatch, NoPathFound,
                          NoUniquePathFound, IncompatibleSchemaVersions, list_available_versions)
from .fleur_schema_parser_functions import AttributeType

__all__ = [
    'InputSchemaDict', 'OutputSchemaDict', 'schema_dict_version_dispatch', 'AttributeType', 'NoPathFound',
    'NoUniquePathFound', 'IncompatibleSchemaVersions', 'SchemaDict', 'list_available_versions'
]



"""
CLI commands for interacting with the fleur schemas in the masci-tools repository
"""
from .root import cli
import click

import masci_tools
from masci_tools.cmdline.utils import echo
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.xml.converters import convert_str_version_number
from masci_tools.io.fleur_xml import load_inpxml, load_outxml
from masci_tools.io.parsers.fleur import inpxml_parser, outxml_parser
from masci_tools.io.parsers.fleur_schema import InputSchemaDict, OutputSchemaDict, list_available_versions

from pathlib import Path
import os
import sys
import shutil
import tempfile
import tabulate

from lxml import etree

GITLAB_URL = 'https://iffgit.fz-juelich.de'
try:
    import gitlab
except ImportError:
    gitlab = None


@cli.group('fleur-schema')
def fleur_schema():
    """Commands related to the Fleur XML Schemas"""


@fleur_schema.command('add')
@click.argument('schema-file', type=click.Path(path_type=Path, resolve_path=True))
@click.option('--overwrite', is_flag=True, help='Overwrite any existing schema-file')
@click.option('--branch',
              type=str,
              help='If the file does not exist the branch can be specified in the fleur git',
              default='develop')
@click.option('--api-key', type=str, help='API key for access to the Iff Gitlab instance', default='')
@click.option('--test-xml-file',
              type=click.Path(exists=True, path_type=Path, resolve_path=True),
              default=None,
              help='Example xmlfile for this schema version to test the file parser against')
@click.option('--from-git', is_flag=True, help='Add the schema from the fleur git repository')
def add_fleur_schema(schema_file, test_xml_file, overwrite, branch, api_key, from_git):
    """
    Adds a new xml schema file to the folder in
    `masci_tools/io/parsers/fleur_schema`
    corresponding to its version number
    """

    PACKAGE_ROOT = Path(masci_tools.__file__).parent.resolve()

    if not isinstance(schema_file, Path):
        if isinstance(schema_file, bytes):
            schema_file = os.fsdecode(schema_file)
        schema_file = Path(schema_file)

    file_name = schema_file.name
    if file_name not in ('FleurInputSchema.xsd', 'FleurOutputSchema.xsd'):
        echo.echo_critical(
            "The Fleur file schema has to be named either 'FleurInputSchema.xsd' or 'FleurOutputSchema.xsd'")

    tmp_dir = None
    if not schema_file.is_file() or from_git:
        if not schema_file.is_file():
            echo.echo_warning(f'{schema_file} does not exist')

        if not from_git and not click.confirm(f'Do you want to download from the fleur git ({branch})'):
            echo.echo_critical('Cannot add Schema file')
        if gitlab is None:
            echo.echo_critical(
                'Cannot download Schema file. Please install python-gitlab or the cmdline-extras requirements')

        api_key = api_key or os.getenv('IFFGIT_APIKEY', '')
        gl = gitlab.Gitlab(GITLAB_URL, private_token=api_key)

        try:
            gl.auth()
        except gitlab.exceptions.GitlabAuthenticationError:
            echo.echo_critical('I am not authorized to access the fleur repository.\n'
                               'Please set the environment variable IFFGIT_APIKEY to access the gitlab instance')

        groups = gl.groups.list(search='fleur')
        for g in groups:
            if g.name == 'fleur':
                repo_id = [repo.id for repo in g.projects.list(all=True) if repo.name == 'fleur'][0]
                project = gl.projects.get(repo_id)
                break
        echo.echo_info(f'Downloading {file_name} from branch {branch} to {schema_file}')
        tmp_dir = tempfile.mkdtemp()
        schema_file = Path(tmp_dir) / file_name
        with open(schema_file, 'wb') as f:
            project.files.raw(file_path=f'io/xml/{file_name}', ref=branch, streamed=True, action=f.write)
        echo.echo_success('Download successful')

    xmlschema = etree.parse(schema_file)
    xmlschema, _ = clear_xml(xmlschema)

    namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
    schema_version = xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0]

    schema_folder = Path(PACKAGE_ROOT) / Path(f'io/parsers/fleur_schema/{schema_version}')

    if schema_file.name == 'FleurInputSchema.xsd':

        destination = schema_folder / 'FleurInputSchema.xsd'
        input_schema = True
        if destination.is_file() and not overwrite:
            echo.echo_critical(
                f'Input Schema for version {schema_version} already exists. Use overwrite=True to replace the Schema')

        echo.echo_info(f"Copying Input Schema for version '{schema_version}' to: '{destination}'")

    elif schema_file.name == 'FleurOutputSchema.xsd':

        input_schema = False
        destination = schema_folder / 'FleurOutputSchema.xsd'
        if destination.is_file() and not overwrite:
            echo.echo_critical(
                f'Output Schema for version {schema_version} already exists. Use overwrite=True to replace the Schema')

        echo.echo_info(f"Copying Output Schema for version '{schema_version}' to: '{destination}'")

    else:
        echo.echo_critical("Fleur file schema has to be named 'FleurInputSchema.xsd' or 'FleurOutputSchema.xsd'")

    os.makedirs(schema_folder, exist_ok=True)
    shutil.copy(schema_file, destination)

    echo.echo_success('Copied Schema file to masci-tools repository')
    if tmp_dir is not None:
        shutil.rmtree(tmp_dir)

    #Make sure that construction of SchemaDicts works for this schema
    if input_schema:
        schema_dict = InputSchemaDict.fromVersion(schema_version, no_cache=True)
    else:
        schema_dict = OutputSchemaDict.fromVersion(schema_version, no_cache=True)
    echo.echo_success('Created Schema dictionary for the given schema file')

    if test_xml_file is not None:

        echo.echo_info(f'Testing Schema for file: {test_xml_file}')
        if input_schema:
            _, schema_dict = load_inpxml(test_xml_file)

            if schema_dict['inp_version'] != schema_version:
                echo.echo_error(
                    f"Test file ({schema_dict['inp_version']}) does not correspond to input version of inserted schema ({schema_version})"
                )
                sys.exit(1)

            parser_info = {}
            parser_dict = inpxml_parser(test_xml_file, parser_info_out=parser_info)
        else:
            _, schema_dict = load_outxml(test_xml_file)

            if schema_dict['out_version'] != schema_version:
                echo.echo_error(
                    f"Test file ({schema_dict['out_version']}) does not correspond to output version of inserted schema ({schema_version})"
                )
                sys.exit(1)

            parser_info = {}
            parser_dict = outxml_parser(test_xml_file, parser_info_out=parser_info)

        echo.echo_info('Parser output:')
        echo.echo_dictionary(parser_dict)
        echo.echo_info('Parser warnings/information:')
        echo.echo_dictionary(parser_info)
        echo.echo_success(f'Parser finished for: {test_xml_file}')


@fleur_schema.command('validate-input')
@click.argument('xml-file', type=click.Path(exists=True))
def validate_inpxmlfile(xml_file):
    """
    Validate the given inp.xml file against the Fleur schema stored for the
    version of the input
    """

    xml_file = os.path.abspath(xml_file)

    xmltree, schema_dict = load_inpxml(xml_file)
    try:
        schema_dict.validate(xmltree)
    except ValueError as err:
        echo.echo_error(str(err))
    else:
        echo.echo_success(f"{xml_file} validates against the schema for version {schema_dict['inp_version']}")


@fleur_schema.command('validate-output')
@click.argument('xml-file', type=click.Path(exists=True))
def validate_outxmlfile(xml_file):
    """
    Validate the given out.xml file against the Fleur schema stored for the
    version of the output
    """
    xml_file = os.path.abspath(xml_file)

    xmltree, schema_dict = load_outxml(xml_file)
    try:
        schema_dict.validate(xmltree)
    except ValueError as err:
        echo.echo_error(str(err))
    else:
        echo.echo_success(f"{xml_file} validates against the schema for version {schema_dict['out_version']}")


@fleur_schema.command('list')
def list_available_schemas():
    """
    Show the available fleur schemas
    """
    input_versions = list_available_versions(output_schema=False)
    output_versions = list_available_versions(output_schema=True)

    all_versions = set(input_versions) | set(output_versions)
    all_versions = sorted(all_versions, key=convert_str_version_number)

    input_versions = [version in input_versions for version in all_versions]
    output_versions = [version in output_versions for version in all_versions]

    echo.echo_info('The following Schemas were found')
    echo.echo(
        tabulate.tabulate(list(zip(all_versions, input_versions, output_versions)),
                          headers=['Version', 'Input Schema available', 'Output Schema available']))



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Common functions for parsing input/output files or XMLschemas from FLEUR
"""
from __future__ import annotations

from masci_tools.util.typing import XMLLike, XPathLike, TXPathLike
from lxml import etree
import warnings
import copy
import logging
from typing import cast

from .xpathbuilder import XPathBuilder


def clear_xml(tree: etree._ElementTree) -> tuple[etree._ElementTree, set[str]]:
    """
    Removes comments and executes xinclude tags of an
    xml tree.

    :param tree: an xml-tree which will be processed

    :returns: cleared_tree, an xmltree without comments and with replaced xinclude tags
    """
    cleared_tree = copy.deepcopy(tree)

    #Remove comments outside the root element (Since they have no parents this would lead to a crash)
    root = cleared_tree.getroot()
    prev_sibling = root.getprevious()
    while prev_sibling is not None:
        if prev_sibling.tag is etree.Comment:
            root.append(prev_sibling)
            root.remove(prev_sibling)
        prev_sibling = prev_sibling.getprevious()

    next_sibling = root.getnext()
    while next_sibling is not None:
        if next_sibling.tag is etree.Comment:
            root.append(next_sibling)
            root.remove(next_sibling)
        next_sibling = next_sibling.getnext()

    #find any include tags
    include_tags: list[etree._Element] = eval_xpath(cleared_tree,
                                                    '//xi:include',
                                                    namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                                                    list_return=True)  #type:ignore

    parents = []
    known_tags = []
    for tag in include_tags:
        parent = tag.getparent()
        if parent is None:
            raise ValueError('Could not find parent of included tag')
        parents.append(parent)
        known_tags.append({elem.tag for elem in parent if isinstance(elem.tag, str)})

    # replace XInclude parts to validate against schema
    if len(include_tags) != 0:
        cleared_tree.xinclude()

    all_included_tags: set[str] = set()
    # get rid of xml:base attribute in the included parts
    for parent, old_tags in zip(parents, known_tags):
        new_tags = {elem.tag for elem in parent if isinstance(elem.tag, str)}

        #determine the elements not in old_tags, which are in tags
        #so what should have been included
        included_tag_names = new_tags.difference(old_tags)

        #Check for empty set (relax.xml include may not insert something)
        if not included_tag_names:
            continue

        all_included_tags = all_included_tags.union(included_tag_names)
        for tag_name in included_tag_names:
            for elem in parent.iterchildren(tag=tag_name):
                for attribute in elem.attrib.keys():
                    if 'base' in attribute:
                        elem.attrib.pop(attribute, '')

    # remove comments from inp.xml
    comments: list[etree._Element] = cleared_tree.xpath('//comment()')  #type:ignore
    for comment in comments:
        com_parent = comment.getparent()
        if com_parent is None:
            raise ValueError('Could not find parent of comment tag')
        com_parent.remove(comment)

    etree.indent(cleared_tree)

    return cleared_tree, all_included_tags


def reverse_xinclude(xmltree, schema_dict, included_tags, **kwargs):
    """
    DEPRECATED ALIAS: Moved to masci_tools.util.schema_dict_util

    Split the xmltree back up according to the given included tags.
    The original xmltree will be returned with the corresponding xinclude tags
    and the included trees are returned in a dict mapping the inserted filename
    to the extracted tree

    Tags for which no known filename is known are returned under unknown-1.xml, ...
    The following tags have known filenames:

        - `relaxation`: ``relax.xml``
        - `kPointLists`: ``kpts.xml``
        - `symmetryOperations`: ``sym.xml``
        - `atomSpecies`: ``species.xml``
        - `atomGroups`: ``atoms.xml``

    Additional mappings can be given in the keyword arguments

    :param xmltree: an xml-tree which will be processed
    :param schema_dict: Schema dictionary containing all the necessary information
    :param included_tags: Iterable of str, containing the names of the tags to be excluded

    :returns: xmltree with the inseerted xinclude tags and a dict mapping the filenames
              to the excluded trees

    :raises ValueError: if the tag can not be found in the given xmltree
    """
    from masci_tools.util.schema_dict_util import reverse_xinclude  #pylint: disable=redefined-outer-name
    warnings.warn('DEPRECATED: reverse_xinclude moved to masci_tools.util.schema_dict_util', DeprecationWarning)
    return reverse_xinclude(xmltree, schema_dict, included_tags, **kwargs)


def validate_xml(xmltree: etree._ElementTree,
                 schema: etree.XMLSchema,
                 error_header: str = 'File does not validate') -> None:
    """
    Checks a given xmltree against a schema and produces a nice error message
    with all the validation errors collected

    :param xmltree: xmltree of the file to validate
    :param schema: etree.XMLSchema to validate against
    :param error_header: str to lead a evtl error message with

    :raises: etree.DocumentInvalid if the schema does not validate
    """
    from itertools import groupby

    try:
        cleared_tree, _ = clear_xml(xmltree)
        schema.assertValid(cleared_tree)
    except etree.DocumentInvalid as exc:
        error_log = sorted(schema.error_log, key=lambda x: x.message)  #type: ignore[call-overload]
        error_output = []
        first_occurence = []
        for message, group in groupby(error_log, key=lambda x: cast(object, x.message)):
            err_occurences = list(group)
            error_message = f'Line {err_occurences[0].line}: {message}'
            error_lines = ''
            if len(err_occurences) > 1:
                error_lines = f"; This error also occurred on the lines {', '.join([str(x.line) for x in err_occurences[1:]])}"
            error_output.append(f'{error_message}{error_lines} \n')
            first_occurence.append(err_occurences[0].line)

        error_output = [line for _, line in sorted(zip(first_occurence, error_output))]
        errmsg = f"{error_header}: \n{''.join(error_output)}"
        raise etree.DocumentInvalid(errmsg) from exc


def eval_xpath(node: XMLLike | etree.XPathElementEvaluator,
               xpath: XPathLike,
               logger: logging.Logger | None = None,
               list_return: bool = False,
               namespaces: dict[str, str] | None = None,
               **variables: etree._XPathObject) -> etree._XPathObject:
    """
    Tries to evaluate an xpath expression. If it fails it logs it.
    If a absolute path is given (starting with '/') and the tag of the node
    does not match the root.
    It will try to find the tag in the path and convert it into a relative path

    :param node: root node of an etree
    :param xpath: xpath expression (relative, or absolute)
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param list_return: if True, the returned quantity is always a list even if only one element is in it
    :param namespaces: dict, passed to namespaces argument in xpath call

    :returns: text, attribute or a node list
    """
    if isinstance(xpath, XPathBuilder):
        xpath_str = xpath.path
        variables = {**variables, **xpath.path_variables}
        xpath = xpath_str

    if logger is not None:
        logger.debug('XPath: %s', xpath)
        logger.debug('XPath Variables: %s', variables)

    if not isinstance(node, (etree._Element, etree._ElementTree, etree.XPathElementEvaluator)):
        if logger is not None:
            logger.error('Wrong Type for xpath eval; Got: %s', type(node))
        raise TypeError(f'Wrong Type for xpath eval; Got: {type(node)}')

    if namespaces is not None and (isinstance(xpath, etree.XPath) or isinstance(node, etree.XPathElementEvaluator)):
        if logger is not None:
            logger.exception(
                'Passing namespaces is only supported for string xpaths and nodes. for etree.XPath or XPathEvaluatore use namespaces in the init function'
            )
        raise ValueError(
            'Passing namespaces is only supported for string xpaths and nodes. for etree.XPath or XPathEvaluatore use namespaces in the init function'
        )

    try:
        if isinstance(node, etree.XPathElementEvaluator):
            if isinstance(xpath, etree.XPath):
                if logger is not None:
                    logger.error('Got an XPath object and an XPathEvaluator in eval_xpath')
                raise TypeError('Got an XPath object and an XPathEvaluator in eval_xpath')
            return_value = node(xpath, **variables)  #[arg-type]
        elif isinstance(xpath, etree.XPath):
            return_value = xpath(node, **variables)
        else:
            return_value = node.xpath(xpath, namespaces=namespaces, smart_strings=True, extensions=None, **variables)
    except etree.XPathEvalError as err:
        if logger is not None:
            logger.exception(
                'There was a XpathEvalError on the xpath: %s \n'
                'The following variables were passed: %s \n'
                'Either it does not exist, or something is wrong with the expression.', xpath, variables)
        raise ValueError(f'There was a XpathEvalError on the xpath: {str(xpath)} \n'
                         f'The following variables were passed: {variables} \n'
                         'Either it does not exist, or something is wrong with the expression.') from err

    if logger is not None:
        logger.debug('XPath Result: %s', return_value)

    if isinstance(return_value, list):
        if len(return_value) == 1 and not list_return:
            return return_value[0]  #type:ignore
    return return_value


def get_xml_attribute(node: etree._Element, attributename: str, logger: logging.Logger | None = None) -> str | None:
    """
    Get an attribute value from a node.

    :param node: a node from etree
    :param attributename: a string with the attribute name.
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :returns: either attributevalue, or None
    """

    if etree.iselement(node):
        attrib_value = node.get(attributename)
        if attrib_value:
            return attrib_value
        if logger is None:
            raise ValueError(f'Tried to get attribute: "{attributename}" from element {node.tag}.\n '
                             f'I received "{attrib_value}", maybe the attribute does not exist')
        logger.warning(
            'Tried to get attribute: "%s" from element %s.\n '
            'I received "%s", maybe the attribute does not exist', attributename, node.tag, attrib_value)

    else:  # something doesn't work here, some nodes get through here
        if logger is None:
            raise TypeError(f'Can not get attributename: "{attributename}" from node of type {type(node)}, '
                            f'because node is not an element of etree.')
        logger.error(
            'Can not get attributename: "%s" from node of type %s, '
            'because node is not an element of etree.', attributename, type(node))

    return None


def split_off_tag(xpath: TXPathLike) -> tuple[TXPathLike, str]:
    """
    Splits off the last part of the given xpath

    .. note::
        etree.XPath objects could lose context in here, i.e.
        non-default options passed at init

    :param xpath:  xpath to split up
    """
    if isinstance(xpath, XPathBuilder):
        xpath = copy.deepcopy(xpath)  #type:ignore[assignment]
        tag = xpath.strip_off_tag()
        return xpath, tag  #type:ignore[return-value]

    if isinstance(xpath, etree.XPath):
        xpath_str = xpath.path
    else:
        xpath_str = xpath  #type:ignore[assignment]

    split_xpath = xpath_str.split('/')
    if split_xpath[-1] == '':
        xpath_str, tag = '/'.join(split_xpath[:-2]), split_xpath[-2]
    else:
        xpath_str, tag = '/'.join(split_xpath[:-1]), split_xpath[-1]

    if isinstance(xpath, etree.XPath):
        xpath = etree.XPath(xpath_str)  #type:ignore [assignment]
    else:
        xpath = xpath_str  #type:ignore[assignment]

    return xpath, tag


def add_tag(xpath: TXPathLike, tag: str) -> TXPathLike:
    """
    Add tag to xpath

    .. note::
        etree.XPath objects could lose context in here, i.e.
        non-default options passed at init

    :param xpath: xpath to change
    :param tag: str of the tag to add

    :returns: xpath with the form {old_xpath}/tag
    """
    if isinstance(xpath, XPathBuilder):
        xpath = copy.deepcopy(xpath)  #type:ignore[assignment]
        xpath.append_tag(tag)
    elif isinstance(xpath, etree.XPath):
        xpath = etree.XPath(f'{str(xpath.path)}/{tag}')  #type:ignore [assignment]
    else:
        xpath = f"{str(xpath).rstrip('/')}/{tag}"  #type:ignore[assignment]
    return xpath


def split_off_attrib(xpath: TXPathLike) -> tuple[TXPathLike, str]:
    """
    Splits off attribute of the given xpath (part after @)

    .. note::
        etree.XPath objects could lose context in here, i.e.
        non-default options passed at init

    :param xpath: xpath to split up
    """
    if isinstance(xpath, XPathBuilder):
        xpath = copy.deepcopy(xpath)  #type:ignore[assignment]
        attrib = xpath.strip_off_tag()
        if '@' not in attrib:
            raise ValueError('Path does not end with an attribute')
        return xpath, attrib.lstrip('@')  #type:ignore[return-value]

    if isinstance(xpath, etree.XPath):
        xpath_str = xpath.path
    else:
        xpath_str = xpath  #type:ignore[assignment]

    split_xpath = xpath_str.split('/@')
    if len(split_xpath) != 2:
        raise ValueError(f"Splitting off attribute failed for: '{split_xpath}'")
    xpath_str, attrib = tuple(split_xpath)

    if isinstance(xpath, etree.XPath):
        xpath = etree.XPath(xpath_str)  #type:ignore [assignment]
    else:
        xpath = xpath_str  #type:ignore[assignment]

    return xpath, attrib


def check_complex_xpath(node: XMLLike | etree.XPathElementEvaluator, base_xpath: XPathLike,
                        complex_xpath: XPathLike) -> None:
    """
    Check that the given complex xpath produces a subset of the results
    for the simple xpath

    :param node: root node of an etree or an etree
    :param base_xpath: str of the xpath without complex syntax
    :param complex_xpath: str of the xpath to check

    :raises ValueError: If the complex_xpath does not produce a subset of the results
                        of the base_xpath
    """
    results_base = set(eval_xpath(node, base_xpath, list_return=True))  #type:ignore
    results_complex = set(eval_xpath(node, complex_xpath, list_return=True))  #type:ignore

    if not results_base.issuperset(results_complex):
        raise ValueError(f"Complex xpath '{complex_xpath!r}' is not compatible with the base_xpath '{base_xpath!r}'")


def abs_to_rel_xpath(xpath: str, new_root: str) -> str:
    """
    Convert a given xpath to be relative from a tag appearing in the
    original xpath.

    :param xpath: str of the xpath to convert
    :param new_root: str of the tag from which the new xpath should be relative

    :returns: str of the relative xpath
    """
    if contains_tag(xpath, new_root):
        xpath = xpath + '/'
        xpath_to_root = '/'.join(xpath.split(new_root + '/')[:-1]) + new_root
        xpath = xpath.replace(f'{xpath_to_root}/', './')
        xpath = xpath.rstrip('/')
    else:
        raise ValueError(f'New root element {new_root} does not appear in xpath {xpath}')

    return xpath


def normalize_xmllike(xmllike: XMLLike) -> etree._Element:
    """
    Returns the root of the xmltree
    """
    if etree.iselement(xmllike):
        return xmllike
    xmllike, _ = clear_xml(xmllike)  #type:ignore[arg-type]
    return xmllike.getroot()


def contains_tag(xpath: XPathLike, tag: str) -> bool:
    """
    Return whether a given xpath contains a given tag
    This assumes that predicates of xpaths can't be nested
    since otherwise the regex for removing them could fail

    This function will only return True if one of the
    tags exactly matches the tag argument not if one tag contains the
    given name in it's name

    :param xpath: xpath expression
    :param tag: tag to check for

    :returns: whether a tag is contained in the xpath
    """
    import re
    if isinstance(xpath, XPathBuilder):
        return tag in xpath.components

    if isinstance(xpath, etree.XPath):
        xpath_str = xpath.path
    else:
        xpath_str = str(xpath)

    #Strip out predicates
    xpath_str = re.sub(r'[\[].*?[\]]', '', xpath_str)
    return tag in xpath_str.split('/')


def is_valid_tag(tag: str) -> bool:
    """
    Return whether the given string is a valid XML tag name

    :param tag: tag to check
    """
    try:
        etree.QName(tag)
        return True
    except ValueError:
        return False



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines decorators for the _TaskParser class to make extending/modifying the parser
more convenient

Up till now 3 decorators are defined:
    - ```register_migration``` marks a function of making backwards incompatible changes
      to the parsing tasks
    - ```register_parsing_function``` gives a mappimg between available parsing functions
      and the keywords in the parsing tasks
    - ```conversion_function``` makes the decorated function available to be called easily
      after a certain parsing task has occurred
"""
from masci_tools.io.parsers.fleur import conversion_function, register_migration
import warnings

warnings.warn('The decorators for the outxml_parser were moved to masci_tools.io.parsers.fleur', DeprecationWarning)



"""
This module is only here for backwards compatibility
"""
import warnings

warnings.warn('The fleur_schema module was moved from masci_tools.io.parsers.fleur'
              'to masci_tools.io.parsers', DeprecationWarning)

from masci_tools.io.parsers.fleur_schema import *



"""
This module provides a place for registering click commands from the tools subpackage
into the main cli
"""
from .root import cli
from masci_tools.tools.fleur_inpxml_converter import inpxml

cli.add_command(inpxml)



# pylint: disable=unused-import
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This subpackage contains transformers for the tabulator subpackage, which turns
properties of a collections of objects into a table.

Transformers let you transform properties while they get tabulated.
"""

import abc as _abc
import typing as _typing
import dataclasses as _dc


@_dc.dataclass(init=True, repr=True, eq=True, order=False, frozen=False)
class TransformedValue:
    """Return type of the :py:class:`~.Transformer`."""
    is_transformed: bool = False
    value: _typing.Union[object, dict] = None
    dtypes: _typing.Union[object, dict] = None
    error: _typing.Optional[Exception] = None


class Transformer(_abc.ABC):
    """Specify how to transformer an object's properties for use in :py:class:`Tabulator`.

    To subclass, you have to implement the :py:meth:`~transformer` method.

    TODO: increase memory performance:

    The following points are meant for transformers using the aiida-jutools `NodeTabulator` implementation of
    Tabulator, but may be of interest  for other implementations. See also TODO in Tabulator docstring.

    - Add documentation for the as yet unmentioned TransformedValue member 'dtypes': a dict of same shape as
      'value' ({new_name: transformed_value}), but with its dict values being the desired dtypes string. This
      is optional, otherwise Tabulator will use standard dtypes or try to guess best dtypes for data on its own.
    """

    @_abc.abstractmethod
    def transform(self,
                  keypath: _typing.Union[str, _typing.List[str]],
                  value: _typing.Any,
                  obj: _typing.Any = None,
                  **kwargs) -> TransformedValue:
        """Specify how to transform properties, based on their keypath and type.

        Extends :py:meth:`~.Transformer.transform`. See also its docstring.

        This default transformer returns all property values unchanged, and so has no effect. To define
        transformations, create a subclass and overwrite with a custom transform method.

         Example: Say, a nested dictionary is passed. It has a property
         `a_dict:{outputs:{last_calc_output_parameters:{attributes:{total_charge_per_atom:[...]`, which
         is a numerical list. We would like the list, and its maximum tabulated as individual columns.
         All other properties included in the tabulation shall be not transformed. We would write
         that transformation rule like this.

        .. code-block:: python

           if keypath == ['outputs', 'last_calc_output_parameters', 'total_charge_per_atom']:
               # assert isinstance(value, list) # optional
               return TransformedValue(True,
                                       {'total_charge_per_atom': value,
                                       'maximum_total_charge': max(value)},
                                       None)

           return TransformedValue(False, value, None)

        All kinds of transformation rules for all kinds of properties can be tailored in this way
        to the specific use-case. Keep in mind that if a include list is used, the property (path) has
        to be included in the include list.

        The keyword arguments `**kwargs` can be used to pass additional arguments, such as external functions, say,
        like `func = external_transform_func`.

        If used in `aiida-jutools`: For accessing process node inputs and outputs Dict nodes properties:
        first key in keypath is 'inputs' or 'outputs', the second is the input or output name that `node.outputs.`
        displays on tab completion (which, under the hood, comes from the in- or outgoing link_label).

        :param keypath: A list of keys, one per nesting level. First key in keypath is usually the object's attribute
        name.
        :param value: The value of the current property.
        :param obj: Optionally, the object containing the property can be passed along. This enables to
                    transform the current property value in combination with other property values.
        :param kwargs: Additional keyword arguments for subclasses.
        :return: A tuple (transformed_value:object, with_new_columns flag:bool). If the flag is False, this
                 means the transformed output property has the same name as the input property (in/out referring
                 here to the input/output of the transform method, not the input/output properties of the object). If
                 the value is a dict, and the flag is set to True, this is understood such that new output
                 properties were created from the input property, and the output value should be interpreted as
                 {property_name: property_value}, possibly with one being the original property name. In a tabulator,
                 new columns would be created for these new properties.
        """


class DefaultTransformer(Transformer):
    """Extends :py:class:`~Transformer`.

    This default transformer does nothing (invariant operation): it just returns the value it received, along with
    the new columns flag with value 'False'.
    """

    def transform(self,
                  keypath: _typing.Union[str, _typing.List[str]],
                  value: _typing.Any,
                  obj: _typing.Any = None,
                  **kwargs) -> _typing.Tuple[_typing.Union[None, _typing.Any, dict], bool]:
        return TransformedValue(is_transformed=False, value=value, error=None)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains useful methods for initializing or modifying a n_mmp_mat file
for LDA+U
"""
from __future__ import annotations
from collections import defaultdict
from typing import NamedTuple

import numpy as np

from masci_tools.io.parsers import fleur_schema
from masci_tools.util.xml.xpathbuilder import XPathBuilder, FilterType
from masci_tools.util.xml.common_functions import get_xml_attribute
from masci_tools.util.xml.common_functions import eval_xpath
from masci_tools.util.typing import XMLLike

from masci_tools.util.schema_dict_util import eval_simple_xpath, evaluate_attribute
from masci_tools.util.schema_dict_util import attrib_exists

from masci_tools.io.io_nmmpmat import write_nmmpmat, write_nmmpmat_from_states, write_nmmpmat_from_orbitals
from masci_tools.io.io_nmmpmat import read_nmmpmat_block, rotate_nmmpmat_block, format_nmmpmat

LINES_PER_BLOCK = 14


def set_nmmpmat(xmltree: XMLLike,
                nmmplines: list[str] | None,
                schema_dict: fleur_schema.SchemaDict,
                species_name: str,
                orbital: int,
                spin: int,
                state_occupations: list[float] | None = None,
                orbital_occupations: list[float] | None = None,
                denmat: np.ndarray | None = None,
                phi: float | None = None,
                theta: float | None = None,
                inverse: bool = False,
                align_to_sqa: bool = False,
                filters: FilterType | None = None) -> list[str]:
    """Routine sets the block in the n_mmp_mat file specified by species_name, orbital and spin
    to the desired density matrix

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the species you want to change
    :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
    :param spin: integer, specifies which spin block should be modified
    :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                              else to zero
    :param denmat: matrix, specify the density matrix explicitly
    :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
    :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :raises ValueError: If something in the input is wrong
    :raises KeyError: If no LDA+U procedure is found on a species

    :returns: list with modified nmmplines
    """

    #All lda+U procedures have to be considered since we need to keep the order
    species_name_base_path = schema_dict.attrib_xpath('name', tag_name='species')
    species_name_xpath = XPathBuilder(species_name_base_path, filters=filters, strict=True)

    if species_name[:4] == 'all-':  #format all-<string>
        species_name_xpath.add_filter('species', {'name': {'contains': species_name[4:]}})
    elif species_name != 'all':
        species_name_xpath.add_filter('species', {'name': species_name})

    possible_species: set[str] = set(eval_xpath(xmltree, species_name_xpath, list_return=True))  #type:ignore

    nspins = _get_number_of_spin_blocks(xmltree, schema_dict)
    if spin > nspins:
        raise ValueError(f'Invalid input: spin {spin} requested, but input has only {nspins} spins')

    if align_to_sqa and (theta is not None or phi is not None):
        raise ValueError(
            'Invalid input: Provide either rotation angles theta or phi or the align_to_sqa switch. Not both')

    ldau_order = _get_ldau_order(xmltree, schema_dict)
    _check_nmmpmat_num_rows(nmmplines, len(ldau_order), nspins, 'set_nmmpmat')

    if state_occupations is not None:
        new_nmmpmat_entry = write_nmmpmat_from_states(orbital, state_occupations, phi=phi, theta=theta, inverse=inverse)
    elif orbital_occupations is not None:
        new_nmmpmat_entry = write_nmmpmat_from_orbitals(orbital,
                                                        orbital_occupations,
                                                        phi=phi,
                                                        theta=theta,
                                                        inverse=inverse)
    elif denmat is not None:
        new_nmmpmat_entry = write_nmmpmat(orbital, denmat, phi=phi, theta=theta, inverse=inverse)
    else:
        raise ValueError('Invalid definition of density matrix. Provide either state_occupations, '
                         'orbital_occupations or denmat')

    for ldau_index, entry in enumerate(ldau_order):
        if entry.species not in possible_species or entry.orbital != orbital:
            continue

        #check if fleurinp has a specified n_mmp_mat file if not initialize it with 0
        if nmmplines is None:
            nmmplines = []
            for _ in range(nspins * LINES_PER_BLOCK * len(ldau_order)):
                nmmplines.append(''.join(map(str, [f'{0.0:20.13f}' for _ in range(7)])))

        #Select the right block from n_mmp_mat and overwrite it with denmatpad
        startRow = ((spin - 1) * len(ldau_order) + ldau_index) * LINES_PER_BLOCK

        nmmplines[startRow:startRow + LINES_PER_BLOCK] = new_nmmpmat_entry

    if nmmplines is None:
        raise ValueError('No denmat blocks set. Probably the species and orbital combination does not exist')

    if align_to_sqa:
        nmmplines = align_nmmpmat_to_sqa(xmltree, nmmplines, schema_dict, species_name, orbital)

    return nmmplines


def align_nmmpmat_to_sqa(xmltree: XMLLike,
                         nmmplines: list[str],
                         schema_dict: fleur_schema.SchemaDict,
                         species_name: str = 'all',
                         orbital: int | str = 'all',
                         phi_before: float | list[float] = 0.0,
                         theta_before: float | list[float] = 0.0,
                         filters: FilterType | None = None) -> list[str]:
    """
    Align the density matrix with the given SQA of the associated species

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the species you want to change
    :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
    :param phi_before: float or list of floats, angle (radian),
                       values for phi for the previous alignment of the density matrix
    :param theta_before: float or list of floats, angle (radian),
                         values for theta for the previous alignment of the density matrix
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :raises ValueError: If something in the input is wrong
    :raises KeyError: If no LDA+U procedure is found on a species

    :returns: list with modified nmmplines
    """
    if nmmplines is None:
        raise ValueError('align_nmmpmat_to_sqa has to be called with a initialized density matrix')

    #All lda+U procedures have to be considered since we need to keep the order
    species_name_base_path = schema_dict.attrib_xpath('name', tag_name='species')
    species_name_xpath = XPathBuilder(species_name_base_path, filters=filters, strict=True)

    if species_name[:4] == 'all-':  #format all-<string>
        species_name_xpath.add_filter('species', {'name': {'contains': species_name[4:]}})
    elif species_name != 'all':
        species_name_xpath.add_filter('species', {'name': species_name})

    possible_species: set[str] = set(eval_xpath(xmltree, species_name_xpath, list_return=True))  #type:ignore

    #Extract the SQA for all atom groups
    # (if we have scond variationn SOC it will just set to the same value)
    all_groups = eval_simple_xpath(xmltree, schema_dict, 'atomgroup', list_return=True)
    sqa_per_group: list[tuple[float, float]] = []
    noco = evaluate_attribute(xmltree, schema_dict, 'l_noco')
    if noco is None:
        noco = False
    soc = evaluate_attribute(xmltree, schema_dict, 'l_soc')
    if soc is None:
        soc = False

    if not isinstance(theta_before, list):
        theta_before = [theta_before] * len(all_groups)
    if not isinstance(phi_before, list):
        phi_before = [phi_before] * len(all_groups)

    if len(theta_before) != len(all_groups) or len(phi_before) != len(all_groups):
        raise ValueError('Not the right number of previous SQA given')
    sqa_before = list(zip(theta_before, phi_before))

    if not noco and not soc:
        raise ValueError('No Spin Quantization axis to align to')

    if not noco and soc:
        theta = evaluate_attribute(xmltree, schema_dict, 'theta', contains='soc')
        phi = evaluate_attribute(xmltree, schema_dict, 'phi', contains='soc')
        sqa_per_group = [(theta, phi)] * len(all_groups)
    else:
        for group in all_groups:
            beta = evaluate_attribute(group, schema_dict, 'beta', contains='noco')
            alpha = evaluate_attribute(group, schema_dict, 'alpha', contains='noco')
            sqa_per_group.append((beta, alpha))

    nspins = _get_number_of_spin_blocks(xmltree, schema_dict)
    ldau_order = _get_ldau_order(xmltree, schema_dict)
    _check_nmmpmat_num_rows(nmmplines, len(ldau_order), nspins, 'align_nmmpmat_to_sqa')

    for ldau_index, entry in enumerate(ldau_order):
        if entry.species not in possible_species or orbital not in (entry.orbital, 'all'):
            continue

        theta, phi = sqa_per_group[entry.group_index]
        theta_before, phi_before = sqa_before[entry.group_index]

        theta -= theta_before
        phi -= phi_before

        for spin in range(nspins):

            startRow = (spin * len(ldau_order) + ldau_index) * LINES_PER_BLOCK
            denmat = read_nmmpmat_block(nmmplines, spin * len(ldau_order) + ldau_index)
            denmat = rotate_nmmpmat_block(denmat, entry.orbital, phi=phi, theta=theta, inverse=True)

            nmmplines[startRow:startRow + LINES_PER_BLOCK] = format_nmmpmat(denmat)

    return nmmplines


def rotate_nmmpmat(xmltree: XMLLike,
                   nmmplines: list[str],
                   schema_dict: fleur_schema.SchemaDict,
                   species_name: str,
                   orbital: int | str,
                   phi: float,
                   theta: float,
                   inverse: bool = False,
                   filters: FilterType | None = None) -> list[str]:
    """
    Rotate the density matrix with the given angles phi and theta

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the species you want to change
    :param orbital: integer or string ('all'), orbital quantum number of the LDA+U procedure to be modified
    :param phi: float, angle (radian), by which to rotate the density matrix
    :param theta: float, angle (radian), by which to rotate the density matrix
    :param filters: Dict specifying constraints to apply on the xpath.
                    See :py:class:`~masci_tools.util.xml.xpathbuilder.XPathBuilder` for details

    :raises ValueError: If something in the input is wrong
    :raises KeyError: If no LDA+U procedure is found on a species

    :returns: list with modified nmmplines
    """
    if nmmplines is None:
        raise ValueError('rotate_nmmpmat has to be called with a initialized density matrix')

    #All lda+U procedures have to be considered since we need to keep the order
    species_name_base_path = schema_dict.attrib_xpath('name', tag_name='species')
    species_name_xpath = XPathBuilder(species_name_base_path, filters=filters, strict=True)

    if species_name[:4] == 'all-':  #format all-<string>
        species_name_xpath.add_filter('species', {'name': {'contains': species_name[4:]}})
    elif species_name != 'all':
        species_name_xpath.add_filter('species', {'name': species_name})

    possible_species: set[str] = set(eval_xpath(xmltree, species_name_xpath, list_return=True))  #type:ignore

    nspins = _get_number_of_spin_blocks(xmltree, schema_dict)
    ldau_order = _get_ldau_order(xmltree, schema_dict)
    _check_nmmpmat_num_rows(nmmplines, len(ldau_order), nspins, 'rotate_nmmpmat')

    for ldau_index, entry in enumerate(ldau_order):
        if entry.species not in possible_species or orbital not in (entry.orbital, 'all'):
            continue

        for spin in range(nspins):

            startRow = (spin * len(ldau_order) + ldau_index) * LINES_PER_BLOCK
            denmat = read_nmmpmat_block(nmmplines, spin * len(ldau_order) + ldau_index)
            denmat = rotate_nmmpmat_block(denmat, entry.orbital, phi=phi, theta=theta, inverse=inverse)

            nmmplines[startRow:startRow + LINES_PER_BLOCK] = format_nmmpmat(denmat)

    return nmmplines


def validate_nmmpmat(xmltree: XMLLike, nmmplines: list[str] | None, schema_dict: fleur_schema.SchemaDict) -> None:
    """
    Checks that the given nmmp_lines is valid with the given xmltree

    Checks that the number of blocks is as expected from the inp.xml and each
    block does not contain non-zero elements outside their size given by the
    orbital quantum number in the inp.xml. Additionally the occupations, i.e.
    diagonal elements are checked that they are in between 0 and the maximum
    possible occupation

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file

    :raises ValueError: if any of the above checks are violated.
    """
    if nmmplines is None:
        return  #Nothing to validate

    nspins = _get_number_of_spin_blocks(xmltree, schema_dict)
    ldau_order = _get_ldau_order(xmltree, schema_dict)
    _check_nmmpmat_num_rows(nmmplines, len(ldau_order), nspins)

    tol = 0.01
    maximum_occupation = 1.0 if nspins > 1 else 2.0

    #Now check for each block if the numbers make sense
    #(no numbers outside the valid area and no nonsensical occupations)
    for ldau_index, entry in enumerate(ldau_order):

        for spin in range(nspins):
            nmmpmat = read_nmmpmat_block(nmmplines, spin * len(ldau_order) + ldau_index)

            #Check for values outside the range -l to l
            outside_val = False
            for index, row in enumerate(nmmpmat):
                if abs(index - 3) > entry.orbital:
                    outside_val = outside_val or any(np.abs(row) > 1e-12)
                else:
                    inside_mask = np.abs(np.array(range(-3, 4))) <= entry.orbital
                    outside_val = outside_val or any(np.abs(row[~inside_mask]) > 1e-12)

            if outside_val:
                raise ValueError(f'Found value outside of valid range in for species {entry.species}, spin {spin+1}'
                                 f' and l={entry.orbital}')

            #check the diagonal for spin-diagonal blocks
            if spin < 2:
                diagonal = nmmpmat.diagonal().real
                invalid_diag = np.logical_or(diagonal < -tol, diagonal > maximum_occupation + tol)

                if invalid_diag.any():
                    raise ValueError(f'Found invalid diagonal element for species {entry.species}, spin {spin+1}'
                                     f' and l={entry.orbital}')


class LDAUElement(NamedTuple):
    """
    Contains the important information needed to locate
    the associated density matrix blocks
    """
    species: str
    orbital: int
    group_index: int


def _get_ldau_order(xmltree: XMLLike, schema_dict: fleur_schema.SchemaDict) -> list[LDAUElement]:
    """
    Get the order of appearance for all LDA+U elements

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    """
    all_ldau = eval_simple_xpath(xmltree, schema_dict, 'ldaU', contains='species', list_return=True)

    species_to_ldauorbital = defaultdict(list)
    for ldau in all_ldau:
        parent = ldau.getparent()
        if parent is None:
            raise ValueError('Could not find parent of tag')
        species_name = get_xml_attribute(parent, 'name')
        orbital = evaluate_attribute(ldau, schema_dict, 'l')
        species_to_ldauorbital[species_name].append(orbital)

    ldau_order: list[LDAUElement] = []
    all_group_species = evaluate_attribute(xmltree, schema_dict, 'species', contains='atomGroup', list_return=True)
    for group_index, species in enumerate(all_group_species):
        if species in species_to_ldauorbital:
            ldau_order.extend(
                LDAUElement(species=species, orbital=orbital, group_index=group_index)
                for orbital in species_to_ldauorbital[species])

    return ldau_order


def _get_number_of_spin_blocks(xmltree: XMLLike, schema_dict: fleur_schema.SchemaDict) -> int:
    """
    Return how many spin blocks should be in the DFT+U density matrix file

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    """

    nspins = evaluate_attribute(xmltree, schema_dict, 'jspins')
    if 'l_mtnocoPot' in schema_dict['attrib_types']:
        if attrib_exists(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
            if evaluate_attribute(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
                nspins = 3
    #yapf: disable
    if schema_dict.inp_version > (0, 32) and attrib_exists(xmltree, schema_dict, 'l_mperp', contains='magnetism'): #type: ignore[attr-defined]
        if evaluate_attribute(xmltree, schema_dict, 'l_mperp', contains='magnetism'):
            nspins = 3
    #yapf: enable

    return nspins


def _check_nmmpmat_num_rows(nmmplines: list[str] | None,
                            num_ldau: int,
                            num_spins: int,
                            name: str | None = None) -> None:
    """Check that number of rows matches the number of lines in nmmp_lines_copy
    If not either there was an n_mmp_mat file present in Fleurinp before and a lda+u calculation
    was added or removed or the n_mmp_mat file was initialized and after the fact lda+u procedures were added
    or removed. In both cases the resolution of this modification is very involved so we throw an error
    """
    if nmmplines is None:
        return

    expected_rows = num_ldau * num_spins * LINES_PER_BLOCK

    #Remove blank lines
    while '' in nmmplines:
        nmmplines.remove('')
    if expected_rows != len(nmmplines):
        hint = ''
        if name is not None:
            hint =  'Either remove the existing file before making modifications ' \
                    f'and only use {name} after all other relevant modifications to the inp.xml'
        raise ValueError(f'The number of lines in n_mmp_mat ({len(nmmplines)}) does not match the number expected from '+\
                         f'the inp.xml file ({expected_rows}). '+hint)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides the classes for easy access to information
from the fleur input and output xsd schema files
"""
from __future__ import annotations

import os
import warnings
import tempfile
import shutil
from functools import update_wrapper, wraps
from pathlib import Path
from typing import Callable, Iterable, TypeVar, Any, cast

from .fleur_schema_parser_functions import TagInfo, convert_str_version_number
try:
    from typing import Literal, Protocol
except ImportError:
    from typing_extensions import Literal, Protocol  #type: ignore

from logging import Logger

from lxml import etree

from masci_tools.util.lockable_containers import LockableDict, LockableList
from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet, CaseInsensitiveDict
from masci_tools.util.xml.common_functions import abs_to_rel_xpath, clear_xml, split_off_tag, contains_tag, validate_xml
from .inpschema_todict import create_inpschema_dict, InputSchemaData
from .outschema_todict import create_outschema_dict, merge_schema_dicts

PACKAGE_DIRECTORY = Path(__file__).parent.resolve()


class NoPathFound(ValueError):
    """
    Exception raised when no path is found for a given tag/attribute
    """


class NoUniquePathFound(ValueError):
    """
    Exception raised when no unique path is found for a given tag/attribute
    """


class IncompatibleSchemaVersions(Exception):
    """
    Exception raised when it is known that a given output version and input version
    cannot be compiled into a complete fleur output xml schema
    """


F = TypeVar('F', bound=Callable[..., Any])
"""Generic Type variable for callable"""


class SchemaDictDispatch(Protocol[F]):
    """Protocol representing function decorated by the schema_dict_version_dispatch decorator"""
    registry: dict[Callable[[tuple[int, int]], bool] | Literal['default'], F]

    def register(self, min_version: str | None = ..., max_version: str | None = ...) -> Callable[[F], F]:
        ...

    dispatch: Callable[[tuple[int, int]], F]
    __call__: F


def schema_dict_version_dispatch(output_schema: bool = False) -> Callable[[F], SchemaDictDispatch[F]]:
    """
    Decorator for creating variations of functions based on the inp/out
    version of the schema_dict. All functions here need to have the signature::

        def f(node, schema_dict, *args, **kwargs):
            pass

    So schema_dict is the second positional argument

    Inspired by singledispatch in the functools module
    """

    def schema_dict_version_dispatch_dec(func: F) -> SchemaDictDispatch[F]:

        registry: dict[Callable[[tuple[int, int]], bool] | Literal['default'], F] = {}

        def dispatch(version: tuple[int, int]) -> F:

            default_match = None
            matches = []
            for condition, func in registry.items():

                if isinstance(condition, str):
                    default_match = func
                elif condition(version):
                    matches.append(func)

            if default_match is None:
                raise ValueError('No default function registered for schema_dict_version dispatch')

            matches.append(default_match)

            if len(matches) > 2:
                raise ValueError('Ambiguous possibilities for schema_dict_version_dispatch for version {version}')

            return matches[0]

        def register(min_version: str | None = None, max_version: str | None = None) -> Callable[[F], F]:

            if min_version is not None:
                min_version_tuple = convert_str_version_number(min_version)

            if max_version is not None:
                max_version_tuple = convert_str_version_number(max_version)

            def register_dec(func: F) -> F:

                if min_version is None and max_version is None:
                    raise ValueError('Either a minimum or maximum version has to be given')

                if min_version is not None and max_version is not None:
                    cond_func = lambda version: min_version_tuple <= version <= max_version_tuple
                elif min_version is not None:
                    cond_func = lambda version: version >= min_version_tuple
                else:
                    cond_func = lambda version: version <= max_version_tuple

                registry[cond_func] = func

                return func

            return register_dec

        @wraps(func)
        def wrapper(node: Any, schema_dict: InputSchemaDict | OutputSchemaDict, *args: Any, **kwargs: Any) -> Any:

            if not isinstance(schema_dict, SchemaDict):
                raise ValueError('Second positional argument is not a SchemaDict')

            if output_schema:
                if not isinstance(schema_dict, OutputSchemaDict):
                    raise ValueError('Second positional argument is not a OutputSchemaDict')
                version = schema_dict.out_version
            else:
                version = schema_dict.inp_version

            return dispatch(version)(node, schema_dict, *args, **kwargs)

        registry['default'] = func
        wrapper.register = register  #type:ignore
        wrapper.dispatch = dispatch  #type:ignore
        wrapper.registry = registry  #type:ignore
        update_wrapper(wrapper, func)

        return cast(SchemaDictDispatch[F], wrapper)

    return schema_dict_version_dispatch_dec


def list_available_versions(output_schema: bool) -> list[str]:
    """
    List the available versions for the schema

    :param output_schema: bool, if True search for FleurOutputSchema.xsd otherwise FleurInputSchema.xsd

    :returns: list version string of the available versions
    """
    versions: list[str] = []
    for f in os.scandir(PACKAGE_DIRECTORY):
        if f.is_dir() and '.' in f.name:
            if output_schema and not (Path(f) / 'FleurOutputSchema.xsd').is_file():
                continue
            if not output_schema and not (Path(f) / 'FleurInputSchema.xsd').is_file():
                continue
            versions.append(f.name)
    return versions


def _get_latest_available_version(output_schema: bool) -> str:
    """
    Determine the newest available version for the schema

    :param output_schema: bool, if True search for FleurOutputSchema.xsd otherwise FleurInputSchema.xsd

    :returns: version string of the latest version
    """
    versions = list_available_versions(output_schema=output_schema)
    return max(versions, key=convert_str_version_number)


def _add_condition(specification: str | Iterable[str] | None, condition: str) -> set[str]:
    """Add element to specification making it into a set if necessary"""

    if specification is None:
        specification = set()
    elif isinstance(specification, str):
        specification = {specification}
    else:
        specification = set(specification)

    specification.add(condition)

    return specification


class SchemaDict(LockableDict):
    """
    Base class for schema dictionaries. Is  locked on initialization with :py:meth:`~masci_tools.util.lockable_containers.LockableDict.freeze()`.
    Holds a reference to the xmlSchema for validating files.

    Also provides interfaces for utility functions

    :param xmlschema: etree.XMLSchema object for validating files

    All other arguments are passed on to :py:class:`~masci_tools.util.lockable_containers.LockableDict`

    """
    _schema_dict_cache: dict[Any, SchemaDict] = {}
    _tag_entries: tuple[str, ...] = ()
    _attrib_entries: tuple[str, ...] = ()
    _info_entries: tuple[str, ...] = ()

    _VALIDATION_ERROR_HEADER: str = 'File does not validate'

    @classmethod
    def clear_cache(cls) -> None:
        """
        Remove all stored entries in the schema dictionary cache
        """
        cls._schema_dict_cache.clear()

    def __init__(self, *args: Any, xmlschema: etree.XMLSchema | None = None, **kwargs: Any):
        if xmlschema is None:
            raise ValueError('xmlschema has to be supplied')
        self.xmlschema = xmlschema
        super().__init__(*args, **kwargs)
        super().freeze()

    def _find_paths(self,
                    name: str,
                    entries: Iterable[str],
                    contains: str | Iterable[str] | None = None,
                    not_contains: str | Iterable[str] | None = None) -> list[str]:
        """
        Find all paths in the schema_dict in the given entries for the given name
        and matching the contains/not_contains criteria

        :param name: str, name of the tag
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path

        :returns: list of str, found xpaths matching the criteria
        """

        if contains is None:
            contains = set()
        elif isinstance(contains, str):
            contains = {contains}
        else:
            contains = set(contains)

        if not_contains is None:
            not_contains = set()
        elif isinstance(not_contains, str):
            not_contains = {not_contains}
        else:
            not_contains = set(not_contains)

        path_list = []
        for entry in entries:
            if name in self[entry]:
                entry_paths = self[entry][name]

                if not isinstance(entry_paths, LockableList):
                    entry_paths = [entry_paths]
                else:
                    entry_paths = entry_paths.get_unlocked()

                invalid_paths = set()
                for phrase in contains:
                    for xpath in entry_paths:
                        if phrase not in xpath:
                            invalid_paths.add(xpath)

                for phrase in not_contains:
                    for xpath in entry_paths:
                        if phrase in xpath:
                            invalid_paths.add(xpath)

                for invalid in invalid_paths:
                    entry_paths.remove(invalid)

                path_list += entry_paths

        return path_list

    def tag_xpath(self,
                  name: str,
                  contains: str | Iterable[str] | None = None,
                  not_contains: str | Iterable[str] | None = None) -> str:
        """
        Tries to find a unique path from the schema_dict based on the given name of the tag
        and additional further specifications

        :param name: str, name of the tag
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path

        :returns: str, xpath for the given tag

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if not self._tag_entries:
            raise NotImplementedError(f"The method 'tag_xpath' cannot be executed for {self.__class__.__name__}"
                                      ' since no tag entries are defined')

        paths = self._find_paths(name, self._tag_entries, contains=contains, not_contains=not_contains)

        if len(paths) == 1:
            return paths[0]
        if len(paths) == 0:
            raise NoPathFound(f'The tag {name} has no possible paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}')
        raise NoUniquePathFound(f'The tag {name} has multiple possible paths with the current specification.\n'
                                f'contains: {contains}, not_contains: {not_contains} \n'
                                f'These are possible: {paths}')

    def relative_tag_xpath(self,
                           name: str,
                           root_tag: str,
                           contains: str | Iterable[str] | None = None,
                           not_contains: str | Iterable[str] | None = None) -> str:
        """
        Tries to find a unique relative path from the schema_dict based on the given name of the tag
        name of the root, from which the path should be relative and additional further specifications

        :param name: str, name of the tag
        :param root_tag: str, name of the tag from which the path should be relative
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path

        :returns: str, xpath for the given tag

        :raises ValueError: If no unique path could be found
        """

        if not self._tag_entries:
            raise NotImplementedError(
                f"The method 'relative_tag_xpath' cannot be executed for {self.__class__.__name__}"
                ' since no tag entries are defined')

        #The paths have to include the root_tag
        contains = _add_condition(contains, root_tag)

        paths = self._find_paths(name, self._tag_entries, contains=contains, not_contains=not_contains)
        #Filter out paths which contain the rootTag explicitly not a tag name containing the root tag
        #e.g. bravaisMatrix vs. bravaisMatrixFilm
        paths = [path for path in paths if contains_tag(path, root_tag)]

        relative_paths = {abs_to_rel_xpath(xpath, root_tag) for xpath in paths}

        if len(relative_paths) == 1:
            return relative_paths.pop()
        if len(relative_paths) == 0:
            raise NoPathFound(f'The tag {name} has no possible relative paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
        raise NoUniquePathFound(f'The tag {name} has multiple possible relative paths with the current specification.\n'
                                f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
                                f'These are possible: {relative_paths}')

    def attrib_xpath(self,
                     name: str,
                     contains: str | Iterable[str] | None = None,
                     not_contains: str | Iterable[str] | None = None,
                     exclude: Iterable[str] | None = None,
                     tag_name: str | None = None) -> str:
        """
        Tries to find a unique path from the schema_dict based on the given name of the attribute
        and additional further specifications

        :param name: str, name of the attribute
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param tag_name: str, if given this name will be used to find a path to a tag with the
                        same name in :py:meth:`tag_xpath()`

        :returns: str, xpath to the tag with the given attribute

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """
        if not self._attrib_entries or not self._info_entries:
            raise NotImplementedError(f"The method 'attrib_xpath' cannot be executed for {self.__class__.__name__}"
                                      ' since no attrib entries are defined')

        if exclude is None:
            exclude = []

        if tag_name is not None:
            tag_xpath = self.tag_xpath(tag_name, contains=contains, not_contains=not_contains)

            tag_info = self.tag_info(
                tag_name,
                contains=contains,
                not_contains=not_contains,
            )

            if name not in tag_info['attribs']:
                raise NoPathFound(f'No attribute {name} found at tag {tag_name}')
            original_case = tag_info['attribs'].original_case[name]
            return f'{tag_xpath}/@{original_case}'

        entries = [entry for entry in self._attrib_entries if all(f'{excl}_attribs' not in entry for excl in exclude)]

        paths = self._find_paths(name, entries, contains=contains, not_contains=not_contains)

        if len(paths) == 1:
            return paths[0]
        if len(paths) == 0:
            raise NoPathFound(f'The attrib {name} has no possible paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}')
        raise NoUniquePathFound(f'The attrib {name} has multiple possible paths with the current specification.\n'
                                f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}\n'
                                f'These are possible: {paths}')

    def relative_attrib_xpath(self,
                              name: str,
                              root_tag: str,
                              contains: str | Iterable[str] | None = None,
                              not_contains: str | Iterable[str] | None = None,
                              exclude: Iterable[str] | None = None,
                              tag_name: str | None = None) -> str:
        """
        Tries to find a unique relative path from the schema_dict based on the given name of the attribute
        name of the root, from which the path should be relative and additional further specifications

        :param schema_dict: dict, containing all the path information and more
        :param name: str, name of the attribute
        :param root_tag: str, name of the tag from which the path should be relative
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param tag_name: str, if given this name will be used to find a path to a tag with the
                        same name in :py:meth:`relative_tag_xpath()`

        :returns: str, xpath for the given tag

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if not self._attrib_entries or not self._info_entries:
            raise NotImplementedError(
                f"The method 'relative_attrib_xpath' cannot be executed for {self.__class__.__name__}"
                ' since no attrib entries are defined')

        if exclude is None:
            exclude = []

        if tag_name is not None:
            tag_xpath = self.relative_tag_xpath(tag_name, root_tag, contains=contains, not_contains=not_contains)

            tag_info = self.tag_info(tag_name, contains=contains, not_contains=not_contains)

            if name not in tag_info['attribs']:
                raise NoPathFound(f'No attribute {name} found at tag {tag_name}')

            original_case = tag_info['attribs'].original_case[name]

            if tag_xpath.endswith('/'):
                return f'{tag_xpath}@{original_case}'
            return f'{tag_xpath}/@{original_case}'

        entries = [entry for entry in self._attrib_entries if all(f'{excl}_attribs' not in entry for excl in exclude)]

        #The paths have to include the root_tag
        contains = _add_condition(contains, root_tag)

        paths = self._find_paths(name, entries, contains=contains, not_contains=not_contains)
        #Filter out paths which contain the rootTag explicitly not a tag name containing the root tag
        #e.g. bravaisMatrix vs. bravaisMatrixFilm
        paths = [path for path in paths if contains_tag(path, root_tag)]
        relative_paths = {abs_to_rel_xpath(xpath, root_tag) for xpath in paths}

        if len(relative_paths) == 1:
            return relative_paths.pop()
        if len(relative_paths) == 0:
            raise NoPathFound(f'The attrib {name} has no possible relative paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
        raise NoUniquePathFound(
            f'The attrib {name} has multiple possible relative paths with the current specification.\n'
            f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
            f'These are possible: {relative_paths}')

    def tag_info(self,
                 name: str,
                 contains: str | Iterable[str] | None = None,
                 not_contains: str | Iterable[str] | None = None,
                 parent: bool = False) -> TagInfo:
        """
        Tries to find a unique path from the schema_dict based on the given name of the tag
        and additional further specifications and returns the tag_info entry for this tag

        :param schema_dict: dict, containing all the path information and more
        :param name: str, name of the tag
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param parent: bool, if True the tag_info for the parent of the tag is returned

        :returns: dict, tag_info for the found xpath
        """

        if not self._tag_entries or not self._info_entries:
            raise NotImplementedError(f"The method 'tag_info' cannot be executed for {self.__class__.__name__}"
                                      ' since no tag or info entries are defined')

        paths = self._find_paths(name, self._tag_entries, contains=contains, not_contains=not_contains)
        if len(paths) == 0:
            raise NoPathFound(f'The tag {name} has no possible paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}')

        EMPTY_TAG_INFO: TagInfo = {
            'attribs': CaseInsensitiveFrozenSet(),
            'optional_attribs': CaseInsensitiveDict(),
            'optional': CaseInsensitiveFrozenSet(),
            'order': [],
            'several': CaseInsensitiveFrozenSet(),
            'simple': CaseInsensitiveFrozenSet(),
            'complex': CaseInsensitiveFrozenSet(),
            'text': CaseInsensitiveFrozenSet()
        }

        tag_info = None
        for path in paths:

            if parent:
                path, _ = split_off_tag(path)

            entry = None
            for info_entry in self._info_entries:
                if path in self[info_entry]:
                    entry = self[info_entry][path]
            if entry is None:
                entry = EMPTY_TAG_INFO

            if tag_info is not None:
                if entry != tag_info:
                    raise NoUniquePathFound(f'Differing tag_info for the found with the current specification\n'
                                            f'contains: {contains}, not_contains: {not_contains}\n'
                                            f'These are possible:  {paths}')
            else:
                tag_info = entry

        if tag_info is None:
            raise ValueError(f'No tag info found for paths: {paths}')

        return tag_info

    def validate(self, xmltree: etree._ElementTree, logger: Logger | None = None, header: str = '') -> None:
        """
        Validate the given XML tree against the schema

        :param xmltree: XML tree to validate
        :param logger: Logger to relay evlt warnings/errors
        """
        header = header or self._VALIDATION_ERROR_HEADER
        errmsg = ''
        xmltree, _ = clear_xml(xmltree)
        try:
            validate_xml(xmltree, self.xmlschema, error_header=header)
        except etree.DocumentInvalid as err:
            errmsg = str(err)
            if logger is not None:
                logger.warning(errmsg)
            raise ValueError(errmsg) from err

        #TODO: Is this an esoteric case that will never happen or is this something we really need?
        if not self.xmlschema.validate(xmltree) and errmsg == '':
            errmsg = f'{header}: Reason is unknown'
            if logger is not None:
                logger.warning(errmsg)
            raise ValueError(errmsg)


class InputSchemaDict(SchemaDict):
    """
    This class contains information parsed from the FleurInputSchema.xsd

    The keys contain the following information:

        :inp_version: Version string of the input schema represented in this object
        :tag_paths: simple xpath expressions to all valid tag names
                    Multiple paths or ambiguous tag names are parsed as a list
        :_basic_types: Parsed definitions of all simple Types with their respective
                       base type (int, float, ...) and evtl. length restrictions
                       (Only used in the schema construction itself)
        :attrib_types: All possible base types for all valid attributes. If multiple are
                       possible a list, with 'string' always last (if possible)
        :simple_elements: All elements with simple types and their type definition
                          with the additional attributes
        :unique_attribs: All attributes and their paths, which occur only once and
                         have a unique path
        :unique_path_attribs: All attributes and their paths, which have a unique path
                              but occur in multiple places
        :other_attribs: All attributes and their paths, which are not in 'unique_attribs' or
                        'unique_path_attribs'
        :omitt_contained_tags: All tags, which only contain a list of one other tag
        :tag_info: For each tag (path), the valid attributes and tags (optional, several,
                   order, simple, text)
    """
    __version__ = '0.2.0'

    _schema_dict_cache: dict[str, InputSchemaDict] = {}  #type:ignore
    _tag_entries = ('tag_paths',)
    _attrib_entries = (
        'unique_attribs',
        'unique_path_attribs',
        'other_attribs',
    )
    _info_entries = ('tag_info',)

    _VALIDATION_ERROR_HEADER: str = 'Input file does not validate against the schema'

    @classmethod
    def fromVersion(cls, version: str, logger: Logger | None = None, no_cache: bool = False) -> InputSchemaDict:
        """
        load the FleurInputSchema dict for the specified version

        :param version: str with the desired version, e.g. '0.33'
        :param logger: logger object for warnings, errors and information, ...

        :return: InputSchemaDict object with the information for the provided version
        """
        schema_file_path = PACKAGE_DIRECTORY / version / 'FleurInputSchema.xsd'

        if not schema_file_path.is_file():

            latest_version = _get_latest_available_version(output_schema=False)

            if int(version.split('.')[1]) < int(latest_version.split('.')[1]):
                message = f'No FleurInputSchema.xsd found at {schema_file_path}'
                raise FileNotFoundError(message)

            if logger is not None:
                logger.warning("No Input Schema available for version '%s'; falling back to '%s'", version,
                               latest_version)
            else:
                warnings.warn(f"No Input Schema available for version '{version}'; falling back to '{latest_version}'")

            version = latest_version
            schema_file_path = PACKAGE_DIRECTORY / version / 'FleurInputSchema.xsd'

        if version in cls._schema_dict_cache and not no_cache:
            return cls._schema_dict_cache[version]

        cls._schema_dict_cache[version] = cls.fromPath(schema_file_path)

        return cls._schema_dict_cache[version]

    @classmethod
    def fromPath(cls, path: os.PathLike) -> InputSchemaDict:
        """
        load the FleurInputSchema dict for the specified FleurInputSchema file

        :param path: path to the input schema file

        :return: InputSchemaDict object with the information for the provided file
        """
        schema_dict = create_inpschema_dict(path)
        xmlschema = etree.XMLSchema(file=path)

        return cls(schema_dict, xmlschema=xmlschema)

    @property
    def inp_version(self) -> tuple[int, int]:
        """
        Returns the input version as an integer for comparisons (`>` or `<`)
        """
        return convert_str_version_number(self.get('inp_version', ''))


class OutputSchemaDict(SchemaDict):
    """
    This object contains information parsed from the FleurOutputSchema.xsd

    The keys contain the following information:

        :out_version: Version string of the output schema represented in this class
        :input_tag: Name of the element containing the fleur input
        :iteration_tags: Names of the elements that can contain all iteration tags
        :tag_paths: simple xpath expressions to all valid tag names not in an iteration
                    Multiple paths or ambiguous tag names are parsed as a list
        :iteration_tag_paths: simple relative xpath expressions to all valid tag names
                              inside an iteration. Multiple paths or ambiguous tag names
                              are parsed as a list
        :_basic_types: Parsed definitions of all simple Types with their respective
                       base type (int, float, ...) and evtl. length restrictions
                       (Only used in the schema construction itself)
        :_input_basic_types: Part of the parsed definitions of all simple Types with their
                             respective base type (int, float, ...) and evtl. length
                             restrictions from the input schema
                             (Only used in the schema construction itself)
        :attrib_types: All possible base types for all valid attributes. If multiple are
                       possible a list, with 'string' always last (if possible)
        :simple_elements: All elements with simple types and their type definition
                          with the additional attributes
        :unique_attribs: All attributes and their paths, which occur only once and
                         have a unique path outside of an iteration
        :unique_path_attribs: All attributes and their paths, which have a unique path
                              but occur in multiple places outside of an iteration
        :other_attribs: All attributes and their paths, which are not in 'unique_attribs' or
                        'unique_path_attribs' outside of an iteration
        :iteration_unique_attribs: All attributes and their relative paths, which occur
                                   only once and have a unique path inside of an iteration
        :iteration_unique_path_attribs: All attributes and their relative paths, which have
                                        a unique path but occur in multiple places inside
                                        of an iteration
        :iteration_other_attribs: All attributes and their relative paths, which are not
                                  in 'unique_attribs' or 'unique_path_attribs' inside
                                  of an iteration
        :omitt_contained_tags: All tags, which only contain a list of one other tag
        :tag_info: For each tag outside of an iteration (path), the valid attributes
                   and tags (optional, several, order, simple, text)
        :iteration_tag_info: For each tag inside of an iteration (relative path),
                             the valid attributes and tags (optional, several,
                             order, simple, text)
    """

    __version__ = '0.2.0'

    _schema_dict_cache: dict[tuple[str, str], OutputSchemaDict] = {}  #type:ignore
    _tag_entries = (
        'tag_paths',
        'iteration_tag_paths',
    )
    _attrib_entries = ('unique_attribs', 'unique_path_attribs', 'other_attribs', 'iteration_unique_attribs',
                       'iteration_unique_path_attribs', 'iteration_other_attribs')
    _info_entries = ('tag_info', 'iteration_tag_info')

    _VALIDATION_ERROR_HEADER: str = 'Output file does not validate against the schema'

    @classmethod
    def fromVersion(cls,
                    version: str,
                    inp_version: str | None = None,
                    logger: Logger | None = None,
                    no_cache: bool = False) -> OutputSchemaDict:
        """
        load the FleurOutputSchema dict for the specified version

        :param version: str with the desired version, e.g. '0.33'
        :param inp_version: str with the desired input version, e.g. '0.33' (defaults to version)
        :param logger: logger object for warnings, errors and information, ...

        :return: OutputSchemaDict object with the information for the provided versions
        """
        if inp_version is None:
            inp_version = version

        schema_file_path = PACKAGE_DIRECTORY / version / 'FleurOutputSchema.xsd'
        inpschema_file_path = PACKAGE_DIRECTORY / inp_version / 'FleurInputSchema.xsd'

        if not schema_file_path.is_file():
            latest_version = _get_latest_available_version(output_schema=True)

            if int(version.split('.')[1]) < int(latest_version.split('.')[1]):
                message = f'No FleurOutputSchema.xsd found at {schema_file_path}'
                raise FileNotFoundError(message)

            if logger is not None:
                logger.warning("No Output Schema available for version '%s'; falling back to '%s'", version,
                               latest_version)
            else:
                warnings.warn(f"No Output Schema available for version '{version}'; falling back to '{latest_version}'")

            version = latest_version
            schema_file_path = PACKAGE_DIRECTORY / version / 'FleurOutputSchema.xsd'

        if not inpschema_file_path.is_file():
            latest_inpversion = _get_latest_available_version(output_schema=False)

            if int(inp_version.split('.')[1]) < int(latest_inpversion.split('.')[1]):
                message = f'No FleurInputSchema.xsd found at {inpschema_file_path}'
                raise FileNotFoundError(message)

            if logger is not None:
                logger.warning("No Input Schema available for version '%s'; falling back to '%s'", inp_version,
                               latest_inpversion)
            else:
                warnings.warn(
                    f"No Input Schema available for version '{inp_version}'; falling back to '{latest_inpversion}'")

            inp_version = latest_inpversion
            inpschema_file_path = PACKAGE_DIRECTORY / version / 'FleurInputSchema.xsd'

        if inp_version != version and logger is not None:
            logger.info('Creating OutputSchemaDict object for differing versions (out: %s; inp: %s)', version,
                        inp_version)

        if (version, inp_version) in cls._schema_dict_cache and not no_cache:
            return cls._schema_dict_cache[(version, inp_version)]

        #Check for known incompatibilities
        if int(version.split('.')[1]) >= 35 and int(inp_version.split('.')[1]) <= 32:
            raise IncompatibleSchemaVersions('Output schemas starting from version 0.35 cannot be compiled '
                                             'to a XML schema with Input schemas before version 0.33')

        inpschema_dict = InputSchemaDict.fromVersion(inp_version, no_cache=no_cache)
        cls._schema_dict_cache[(version, inp_version)] = cls.fromPath(schema_file_path,
                                                                      inp_path=inpschema_file_path,
                                                                      inpschema_dict=inpschema_dict)

        return cls._schema_dict_cache[(version, inp_version)]

    @classmethod
    def fromPath(cls,
                 path: os.PathLike,
                 inp_path: os.PathLike | None = None,
                 inpschema_dict: InputSchemaDict | None = None) -> OutputSchemaDict:
        """
        load the FleurOutputSchema dict for the specified paths

        :param path: path to the FleurOutputSchema file
        :param inp_path: path to the FleurInputSchema file (defaults to same folder as path)

        :return: OutputSchemaDict object with the information for the provided files
        """

        if inp_path is None:
            inp_path = Path(path).parent / 'FleurInputSchema.xsd'

        if inpschema_dict is None:
            inpschema_dict = create_inpschema_dict(inp_path)  #type:ignore
        inpschema_data = cast(InputSchemaData, inpschema_dict)

        schema_dict = create_outschema_dict(path, inpschema_dict=inpschema_data)
        schema_dict = merge_schema_dicts(inpschema_data, schema_dict)

        with tempfile.TemporaryDirectory() as td:
            td_path = Path(td)
            temp_input_schema_path = td_path / 'FleurInputSchema.xsd'
            shutil.copy(inp_path, temp_input_schema_path)

            temp_output_schema_path = td_path / 'FleurOutputSchema.xsd'
            shutil.copy(path, temp_output_schema_path)

            xmlschema = etree.XMLSchema(file=temp_output_schema_path)

        return cls(schema_dict, xmlschema=xmlschema)

    @property
    def inp_version(self) -> tuple[int, int]:
        """
        Returns the input version as an integer for comparisons (`>` or `<`)
        """
        return convert_str_version_number(self.get('inp_version', ''))

    @property
    def out_version(self) -> tuple[int, int]:
        """
        Returns the output version as an integer for comparisons (`>` or `<`)
        """
        return convert_str_version_number(self.get('out_version', ''))

    def iteration_tag_xpath(self,
                            name: str,
                            contains: str | Iterable[str] | None = None,
                            not_contains: str | Iterable[str] | None = None,
                            iteration_tag: str = 'iteration') -> str:
        """
        Tries to find a unique path from the schema_dict based on the given name of the tag
        and additional further specifications in the iteration section of the out.xml and returns
        the absolute path to it

        :param name: str, name of the tag
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param iteration_tag: name of the tag containing the iteration information

        :returns: str, xpath for the given tag

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if iteration_tag not in self['iteration_tags']:
            raise ValueError(f"{iteration_tag} is not a valid iteration tag valid are: {list(self['iteration_tags'])}")
        iteration_path = self.tag_xpath(iteration_tag)

        paths = self._find_paths(name, ('iteration_tag_paths',), contains=contains, not_contains=not_contains)
        paths = [f"{iteration_path}{path.lstrip('.')}" for path in paths]

        if len(paths) == 1:
            return paths[0]
        if len(paths) == 0:
            raise NoPathFound(f'The tag {name} has no possible iteration paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}')
        raise NoUniquePathFound(
            f'The tag {name} has multiple possible iteration paths with the current specification.\n'
            f'contains: {contains}, not_contains: {not_contains} \n'
            f'These are possible: {paths}')

    def relative_iteration_tag_xpath(self,
                                     name: str,
                                     root_tag: str,
                                     contains: str | Iterable[str] | None = None,
                                     not_contains: str | Iterable[str] | None = None,
                                     iteration_tag: str = 'iteration') -> str:
        """
        Tries to find a unique path from the schema_dict based on the given name of the tag
        and additional further specifications in the iteration section of the out.xml and returns
        the absolute path to it

        :param name: str, name of the tag
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param iteration_tag: name of the tag containing the iteration information

        :returns: str, xpath for the given tag

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if iteration_tag not in self['iteration_tags']:
            raise ValueError(f"{iteration_tag} is not a valid iteration tag valid are: {list(self['iteration_tags'])}")
        iteration_path = self.tag_xpath(iteration_tag)

        if not contains_tag(iteration_path, root_tag):
            #The paths have to include the root_tag
            contains = _add_condition(contains, root_tag)

        paths = self._find_paths(name, ('iteration_tag_paths',), contains=contains, not_contains=not_contains)
        if not contains_tag(iteration_path, root_tag):
            #Filter out paths which contain the rootTag explicitly not a tag name containing the root tag
            #e.g. bravaisMatrix vs. bravaisMatrixFilm
            paths = [path for path in paths if contains_tag(path, root_tag)]
        paths = [f"{iteration_path}{path.lstrip('.')}" for path in paths]
        relative_paths = {abs_to_rel_xpath(xpath, root_tag) for xpath in paths}

        if len(paths) == 1:
            return relative_paths.pop()
        if len(paths) == 0:
            raise NoPathFound(
                f'The tag {name} has no possible relative iteration paths with the current specification.\n'
                f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
        raise NoUniquePathFound(
            f'The tag {name} has multiple possible relative iteration paths with the current specification.\n'
            f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
            f'These are possible: {relative_paths}')

    def iteration_attrib_xpath(self,
                               name: str,
                               contains: str | Iterable[str] | None = None,
                               not_contains: str | Iterable[str] | None = None,
                               exclude: Iterable[str] | None = None,
                               tag_name: str | None = None,
                               iteration_tag: str = 'iteration') -> str:
        """
        Tries to find a unique path from the schema_dict based on the given name of the attribute
        and additional further specifications in the iteration section of the out.xml and returns
        the absolute path to it

        :param name: str, name of the attribute
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param tag_name: str, if given this name will be used to find a path to a tag with the
                        same name in :py:meth:`iteration_tag_xpath()`
        :param iteration_tag: name of the tag containing the iteration information

        :returns: str, xpath to the tag with the given attribute

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if not self._attrib_entries or not self._info_entries:
            raise NotImplementedError(
                f"The method 'iteration_attrib_xpath' cannot be executed for {self.__class__.__name__}"
                ' since no attrib entries are defined')

        if iteration_tag not in self['iteration_tags']:
            raise ValueError(f"{iteration_tag} is not a valid iteration tag valid are: {list(self['iteration_tags'])}")
        iteration_path = self.tag_xpath(iteration_tag)

        if exclude is None:
            exclude = []

        if tag_name is not None:
            tag_xpath = self.iteration_tag_xpath(tag_name, contains=contains, not_contains=not_contains)

            tag_info = self.tag_info(
                tag_name,
                contains=contains,
                not_contains=not_contains,
            )

            if name not in tag_info['attribs']:
                raise NoPathFound(f'No attribute {name} found at tag {tag_name}')
            original_case = tag_info['attribs'].original_case[name]
            return f'{tag_xpath}/@{original_case}'

        entries = [
            entry for entry in self._attrib_entries
            if 'iteration' in entry and all(f'{excl}_attribs' not in entry for excl in exclude)
        ]

        paths = self._find_paths(name, entries, contains=contains, not_contains=not_contains)
        paths = [f"{iteration_path}{path.lstrip('.')}" for path in paths]

        if len(paths) == 1:
            return paths[0]
        if len(paths) == 0:
            raise NoPathFound(f'The attrib {name} has no possible iteration paths with the current specification.\n'
                              f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}')
        raise NoUniquePathFound(
            f'The attrib {name} has multiple possible iteration paths with the current specification.\n'
            f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}\n'
            f'These are possible: {paths}')

    def relative_iteration_attrib_xpath(self,
                                        name: str,
                                        root_tag: str,
                                        contains: str | Iterable[str] | None = None,
                                        not_contains: str | Iterable[str] | None = None,
                                        exclude: Iterable[str] | None = None,
                                        tag_name: str | None = None,
                                        iteration_tag: str = 'iteration') -> str:
        """
        Tries to find a unique relative path from the schema_dict based on the given name of the attribute
        name of the root, from which the path should be relative and additional further specifications

        :param schema_dict: dict, containing all the path information and more
        :param name: str, name of the attribute
        :param root_tag: str, name of the tag from which the path should be relative
        :param contains: str or list of str, this string has to be in the final path
        :param not_contains: str or list of str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param tag_name: str, if given this name will be used to find a path to a tag with the
                        same name in :py:meth:`relative_iteration_tag_xpath()`
        :param iteration_tag: name of the tag containing the iteration information

        :returns: str, xpath for the given tag

        :raises NoPathFound: If no path matching the criteria could be found
        :raises NoUniquePathFound: If multiple paths matching the criteria are found
        """

        if not self._attrib_entries or not self._info_entries:
            raise NotImplementedError(
                f"The method 'iteration_relative_attrib_xpath' cannot be executed for {self.__class__.__name__}"
                ' since no attrib entries are defined')

        if iteration_tag not in self['iteration_tags']:
            raise ValueError(f"{iteration_tag} is not a valid iteration tag valid are: {list(self['iteration_tags'])}")
        iteration_path = self.tag_xpath(iteration_tag)

        if exclude is None:
            exclude = []

        if tag_name is not None:
            tag_xpath = self.relative_iteration_tag_xpath(tag_name,
                                                          root_tag,
                                                          contains=contains,
                                                          not_contains=not_contains)

            tag_info = self.tag_info(tag_name, contains=contains, not_contains=not_contains)

            if name not in tag_info['attribs']:
                raise NoPathFound(f'No attribute {name} found at tag {tag_name}')

            original_case = tag_info['attribs'].original_case[name]

            if tag_xpath.endswith('/'):
                return f'{tag_xpath}@{original_case}'
            return f'{tag_xpath}/@{original_case}'

        entries = [
            entry for entry in self._attrib_entries
            if 'iteration' in entry and all(f'{excl}_attribs' not in entry for excl in exclude)
        ]

        if not contains_tag(iteration_path, root_tag):
            contains = _add_condition(contains, root_tag)

        paths = self._find_paths(name, entries, contains=contains, not_contains=not_contains)
        if not contains_tag(iteration_path, root_tag):
            #Filter out paths which contain the rootTag explicitly not a tag name containing the root tag
            #e.g. bravaisMatrix vs. bravaisMatrixFilm
            paths = [path for path in paths if contains_tag(path, root_tag)]
        paths = [f"{iteration_path}{path.lstrip('.')}" for path in paths]
        relative_paths = {abs_to_rel_xpath(xpath, root_tag) for xpath in paths}

        if len(relative_paths) == 1:
            return relative_paths.pop()
        if len(relative_paths) == 0:
            raise NoPathFound(
                f'The attrib {name} has no possible relative iteration paths with the current specification.\n'
                f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
        raise NoUniquePathFound(
            f'The attrib {name} has multiple possible relative iteration paths with the current specification.\n'
            f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
            f'These are possible: {relative_paths}')
